"""
Primary class for performing the parameter fitting in KIPET

"""
# Standard library imports
import copy
import os
import time
import warnings

# Third party imports
import numpy as np
import pandas as pd
from pyomo.environ import (Constraint, ConstraintList, Objective,
    SolverFactory, Suffix, TerminationCondition, value, Var)
from scipy.sparse import coo_matrix
import scipy.stats as st

# KIPET library imports
from kipet.model_components.objectives import (comp_objective,
                                               conc_objective)
from kipet.input_output.read_hessian import split_sipopt_string, read_reduce_hessian, save_eig_red_hess
from kipet.estimator_tools.pyomo_simulator import PyomoSimulator
from kipet.estimator_tools.results_object import ResultsObject
from kipet.mixins.parameter_estimator_mixins import PEMixins
from kipet.model_tools.pyomo_model_tools import convert
from kipet.general_settings.variable_names import VariableNames


class ParameterEstimator(PEMixins, PyomoSimulator):

    """Optimizer for parameter estimation"""

    def __init__(self, model):
        super(ParameterEstimator, self).__init__(model)

        self.__var = VariableNames()

        self.hessian = None
        self.cov = None
        self._estimability = False
        self._idx_to_variable = dict()
        self._n_actual = self._n_components
        self.model_variance = True
        self.termination_condition = None
        
        # This should be a subclass or a mixin
        self.G_contribution = None
        self.unwanted_G = False
        self.time_variant_G = False
        self.time_invariant_G = False
        self.time_invariant_G_decompose = False
        self.time_invariant_G_no_decompose = False
        
        # Parameter name list
        self.param_names = []
        self.param_names_full = [f'P[{p}]' for p in getattr(self.model, self.__var.model_parameter)]
        self.param_names = [p for p in getattr(self.model, 'P')]
        if hasattr(self.model, 'Pinit'):
            self.param_names_full += [f'init_conditions[{p}]' for p in getattr(self.model, 'Pinit')]
            self.param_names += [p for p in getattr(self.model, 'Pinit')]

        if hasattr(self.model, 'non_absorbing'):
            warnings.warn("Overriden by non_absorbing")
            list_components = [k for k in self._mixture_components if k not in self._non_absorbing]
            self._sublist_components = list_components
            self._n_actual = len(self._sublist_components)

        # for new huplc structure (CS):
        if self._huplc_given:
            self._list_huplcabs = self._huplc_absorbing
            # self._n_huplc = len(self._list_huplcabs)

        if hasattr(self.model, 'known_absorbance'):
            warnings.warn("Overriden by known_absorbance")
            list_components = [k for k in self._mixture_components if k not in self._known_absorbance]
            self._sublist_components = list_components
            self._n_actual = len(self._sublist_components)

        else:
            self._sublist_components = [k for k in self._mixture_components]
            
        if hasattr(self, '_abs_components'):
            self.component_set = self._abs_components
            self.component_var = self.__var.concentration_spectra_abs
            self.n_val = self._nabs_components
            
        else:
            self.component_set = self._sublist_components
            self.component_var = self.__var.concentration_spectra
            self.n_val = self._n_actual
            
    def run_opt(self, solver, **kwds):

        """ Solves parameter estimation problem.

        :param str solver: The solver used to solve the NLP
        :param dict kwds: The dictionary of options passed from the ReactionModel

        :Keyword Args:
            - solver (str): name of the nonlinear solver to used
            - solver_opts (dict, optional): options passed to the nonlinear solver
            - variances (dict or float, optional): map of component name to noise variance. The
              map also contains the device noise variance. If not float then we only use device variance
              and ignore model variance.
            - tee (bool,optional): flag to tell the optimizer whether to stream output
              to the terminal or not.
            - with_d_vars (bool,optional): flag to the optimizer whether to add
            - variables and constraints for D_bar(i,j).
            - report_time (bool, optional): flag as to whether to time the parameter estimation or not.
            - estimability (bool, optional): flag to tell the model whether it is
              being used by the estimability analysis and therefore will need to return the
              hessian for analysis.
            - model_variance (bool, optional): Default is True. Flag to tell whether we are only
              considering the variance in the device, or also model noise as well.
            - model_variance (bool, optional): Default is True. Flag to tell whether we are only
              considering the variance in the device, or also model noise as well.

        :return: The results from the parameter fitting
        :rtype: ResultsObject

        """
        run_opt_kwargs = copy.copy(kwds)
        
        solver_opts = kwds.pop('solver_opts', dict())
        variances = kwds.pop('variances', dict())
        tee = kwds.pop('tee', False)
        with_d_vars = kwds.pop('with_d_vars', False)
        covariance = kwds.pop('covariance', False)
 
        estimability = kwds.pop('estimability', False)
        report_time = kwds.pop('report_time', False)
        model_variance = kwds.pop('model_variance', True)

        inputs_sub = kwds.pop("inputs_sub", None)
        trajectories = kwds.pop("trajectories", None)
        fixedtraj = kwds.pop("fixedtraj", False)
        fixedy = kwds.pop("fixedy", False)
        yfix = kwds.pop("yfix", None)
        yfixtraj = kwds.pop("yfixtraj", None)

        jump = kwds.pop("jump", False)
        
        self.confidence = kwds.pop('confidence', None)
        if self.confidence is None:
            self.confidence = 0.95
        
        G_contribution = kwds.pop('G_contribution', None)
        St = kwds.pop('St', dict())
        Z_in = kwds.pop('Z_in', dict())

        self.solver = solver
        self.model_variance = model_variance
        self._estimability = estimability

        if not self.model.alltime.get_discretization_info():
            raise RuntimeError('apply discretization first before initializing')

        self.G_contribution = G_contribution
       
        if report_time:
            start = time.time()

        opt = SolverFactory(self.solver)
        
        if self.G_contribution == 'time_invariant_G':
            self.decompose_G_test(St, Z_in)
        self.g_handling_status_messages()

        if covariance:
            if self.solver != 'ipopt_sens' and self.solver != 'k_aug':
                raise RuntimeError('To get covariance matrix the solver needs to be ipopt_sens or k_aug')
            
            if self.solver == 'ipopt_sens':
                if not 'compute_red_hessian' in solver_opts.keys():
                    solver_opts['compute_red_hessian'] = 'yes'
            
            if self.solver == 'k_aug':
                solver_opts['compute_inv'] = ''

            self._define_reduce_hess_order()

        if inputs_sub is not None:
            from kipet.estimator_tools.additional_inputs import add_inputs
            
            add_kwargs = dict(
                fixedtraj = fixedtraj,
                fixedy = fixedy, 
                inputs_sub = inputs_sub,
                yfix = yfix,
                yfixtraj = yfixtraj,
                trajectories = trajectories,
            )
            add_inputs(self, add_kwargs)
        
        if jump:
            from kipet.estimator_tools.jumps_method import set_up_jumps
            set_up_jumps(self.model, run_opt_kwargs)
            
        for key, val in solver_opts.items():
            opt.options[key] = val
            
        self.objective_value = 0
        active_objectives = [o for o in self.model.component_map(Objective, active=True)]        
        if active_objectives:
            print(
                "WARNING: The model has an active objective. Running optimization with models objective.\n"
                " To solve optimization with default objective (Weifengs) deactivate all objectives in the model.")
            solver_results = opt.solve(self.model, tee=tee)

        elif self._spectra_given:
            self.objective_value, self.cov_mat = self._solve_extended_model(
                variances, 
                opt,
                tee=tee,
                covariance=covariance,
                with_d_vars=with_d_vars,
                **kwds)

        elif self._concentration_given:
            self.objective_value, self.cov_mat = self._solve_model_given_c(
                variances, 
                opt,
                tee=tee,
                covariance=covariance,
                **kwds)
       
        else:
            raise RuntimeError(
                'Must either provide concentration data or spectra in order to solve the parameter estimation problem')

        if report_time:
            end = time.time()
            print("Total execution time in seconds for variance estimation:", end - start)

        return self._get_results()

    def _get_results(self):
        """Removed results unit from function

        :return: The formatted results
        :rtype: ResultsObject

        """
        results = ResultsObject()
        results.objective = self.objective_value
        results.parameter_covariance = self.cov
        results.load_from_pyomo_model(self.model)
        results.show_parameters(self.confidence)

        if self._spectra_given:
            from kipet.calculation_tools.beer_lambert import D_from_SC
            D_from_SC(self.model, results)

        if hasattr(self.model, self.__var.model_parameter_scaled): 
            setattr(results, self.__var.model_parameter, {name: getattr(self.model, self.__var.model_parameter)[name].value*getattr(self.model, self.__var.model_parameter_scaled)[name].value for name in self.model.parameter_names})
        else:
            setattr(results, self.__var.model_parameter, {name: getattr(self.model, self.__var.model_parameter)[name].value for name in self.model.parameter_names})

        if self.termination_condition!=None and self.termination_condition!=TerminationCondition.optimal:
            raise Exception("The current iteration was unsuccessful.")
        else:
            if self._estimability == True:
                return self.hessian, results
            else:
                return results

        return results

    def _get_list_components(self, species_list):
        """Returns the list of components used

        :param list species_list: List of species

        :return: List of model components
        :rtype: list

        """
        if species_list is None:
            list_components = [k for k in self._mixture_components]
        else:
            list_components = []
            for k in species_list:
                if k in self._mixture_components:
                    list_components.append(k)
                else:
                    warnings.warn("Ignored {} since is not a mixture component of the model".format(k))

        return list_components
            
    def _get_objective_expr(self, model, with_d_vars, component_set, sigma_sq, device=False):
        """Build the objective expression for D_bar (primary objective term in KIPET)

        :param ConcreteModel model: The model for use in parameter fitting
        :param bool with_d_vars: For spectral problems
        :param list component_set: A list of the components
        :param dict sigma_sq: A dictionary of component variances
        :param bool device: Indicates if the device variance is used

        :return expr: The objective expression
        :rtype: expression

        """
        expr = 0
        for t in model.times_spectral:
            for l in model.meas_lambdas:
                
                #print(t, l)
                if with_d_vars:
                #    print('Here 1')
                    D_bar = model.D_bar[t, l]
                else:
                    if device:    
                        if hasattr(model, 'huplc_absorbing'):
                #            print('Here 2')
                            D_bar = sum(model.Z[t, k] * model.S[l, k] for k in component_set if k not in model.solid_spec_arg1)
                        else:
                #            print('Here 3')
                            D_bar = sum(model.Z[t, k] * model.S[l, k] for k in component_set)    
                    else:
                        if hasattr(model, '_abs_components'):
                            if hasattr(model, 'huplc_absorbing') and hasattr(model, 'solid_spec_arg1'):
                #                print('Here 4')
                                D_bar = sum(model.Cs[t, k] * model.S[l, k] for k in model._abs_components if k not in model.solid_spec_arg1)
                            else:
                #                print('Here 5')
                                D_bar = sum(model.Cs[t, k] * model.S[l, k] for k in model._abs_components)
                        else:
                            if hasattr(model, 'huplc_absorbing') and hasattr(model, 'solid_spec_arg1'):
                #                print('Here 6')
                                D_bar = sum(model.C[t, k] * model.S[l, k] for k in component_set if k not in model.solid_spec_arg1)
                            else:
                #                print('Here 7')
                                D_bar = sum(model.C[t, k] * model.S[l, k] for k in component_set)
                
                #print(D_bar)
                            
                if self.G_contribution == 'time_variant_G':
                    expr += (model.D[t, l] - D_bar - model.qr[t]*model.g[l]) ** 2 / (sigma_sq['device'])
                elif self.time_invariant_G_no_decompose:
                    expr += (model.D[t, l] - D_bar - model.g[l]) ** 2 / (sigma_sq['device'])
                else:
                    expr += (model.D[t, l] - D_bar) ** 2 / (sigma_sq['device'])
                   
        #print(expr) 
                   
        return expr 

    def _solve_extended_model(self, sigma_sq, optimizer, **kwds):
        """Solves estimation based on spectral data. (known variances)

           This method is not intended to be used by users directly

        :param dict sigma_sq: variances
        :param SolverFactory optimizer: Pyomo Solver factory object

        :Keyword Args:

            - tee (bool,optional): flag to tell the optimizer whether to stream output
              to the terminal or not
            - with_d_vars (bool,optional): flag to the optimizer whether to add
              variables and constraints for D_bar(i,j)
            - subset_lambdas (array_like,optional): Set of wavelengths to used in
              the optimization problem (not yet fully implemented). Default all wavelengths.

        :return: None

        """
        tee = kwds.pop('tee', False)
        with_d_vars = kwds.pop('with_d_vars', False)

        if self._huplc_given:  # added for new huplc structure CS
            weights = kwds.pop('weights', [1.0, 1.0, 1.0])
        else:
            weights = kwds.pop('weights', [1.0, 1.0])
        warmstart = kwds.pop('warmstart', False)
        eigredhess2file = kwds.pop('eigredhess2file', False)
        penaltyparam = kwds.pop('penaltyparam', False)
        penaltyparamcon = kwds.pop('penaltyparamcon', False) #added for optional penalty term related to constraint CS
        ppenalty_dict = kwds.pop('ppenalty_dict', None)
        ppenalty_weights = kwds.pop('ppenalty_weights', None)
        covariance = kwds.pop('covariance', False)
        species_list = kwds.pop('subset_components', None)
        set_A = kwds.pop('subset_lambdas', list())
        
        self._eigredhess2file=eigredhess2file
        cov_mat = None

        if not set_A:
            set_A = self.model.meas_lambdas

        list_components = self._get_list_components(species_list)
        
        if not self._spectra_given:
            raise NotImplementedError("Extended model requires spectral data model.D[ti,lj]")

        if hasattr(self.model, 'non_absorbing'):
            warnings.warn("Overriden by non_absorbing!")
            list_components = [k for k in self._mixture_components if k not in self._non_absorbing]

        if hasattr(self.model, 'known_absorbance'):
            warnings.warn("Overriden by known_absorbance!")
            list_components = [k for k in self._mixture_components if k not in self._known_absorbance]

        all_sigma_specified = True

        if isinstance(sigma_sq, dict): 
            keys = sigma_sq.keys()
            for k in list_components:
                if k not in keys:
                    all_sigma_specified = False
                    sigma_sq[k] = max(sigma_sq.values())

            if not 'device' in sigma_sq.keys():
                all_sigma_specified = False
                sigma_sq['device'] = 1.0

        elif isinstance(sigma_sq, float):
            sigma_dev = sigma_sq
            sigma_sq = dict()
            sigma_sq['device'] = sigma_dev

        model = self.model
        
        # Instead of calling (19), (24) is called but all qr[i] are fixed at 1.0. KH.L
        if self.time_invariant_G_no_decompose:
            for i in model.times_spectral:
                model.qr[i] = 1.0
            model.qr.fix()

        def _qr_end_constraint(model):
            return model.qr[model.times_spectral[-1]] == 1.0
        
        if self.G_contribution == 'time_variant_G':
            model.qr_end_cons = Constraint(rule = _qr_end_constraint)
        
        
        if with_d_vars and self.model_variance:
            model.D_bar = Var(model.times_spectral, model.meas_lambdas)
            
            def rule_D_bar(model, t, l):
                if hasattr(model, 'huplc_absorbing') and hasattr(model, 'solid_spec_arg1'):
                    return model.D_bar[t, l] == sum(
                        getattr(model, self.component_var)[t, k] * model.S[l, k] for k in self.component_set if k not in model.solid_spec_arg1)
                else:
                    return model.D_bar[t, l] == sum(getattr(model, self.component_var)[t, k] * model.S[l, k] for k in self.component_set)

            model.D_bar_constraint = Constraint(model.times_spectral,
                                                model.meas_lambdas,
                                                rule=rule_D_bar)
        
        
        #For addition of huplc data and matching liquid and solid species (CS):
        def rule_Dhat_bar(model, t, l):
            list_huplcabs = [k for k in model.huplc_absorbing.value]
            return model.Dhat_bar[t, l] == (model.Z[t, l] + model.solidvol[t, l]) / (
                sum(model.Z[t, j] + model.solidvol[t, j] for j in list_huplcabs))

        def rule_objective(m):
            
            terms = {}
            obj_weight = {}
            
            terms['D_bar'] = self._get_objective_expr(m, with_d_vars, list_components, sigma_sq)
            obj_weight['D_bar'] = weights[0]

            terms['model_fit'] = 0
            obj_weight['model_fit'] = 1
            
            for t in m.times_spectral:
                terms['model_fit'] += sum((model.C[t, k] - model.Z[t, k]) ** 2 / sigma_sq[k] for k in list_components)
                
            terms['L2_penalty'] = 0
            obj_weight['L2_penalty'] = 1
            # L2 penalty term to objective penalizing values that deviate from values defined in ppenalty_dict:
            if penaltyparam==True:
                if ppenalty_weights is None:
                    terms['L2_penalty'] = 0.0
                    for k in model.P.keys():
                        if k in ppenalty_dict.keys():
                            terms['L2_penalty'] += (model.P[k] - ppenalty_dict[k]) ** 2
                else:
                    if len(ppenalty_dict)!=len(ppenalty_weights):
                        raise RuntimeError(
                            'For every penalty term a weight must be defined.')
                    if ppenalty_dict.keys()!=ppenalty_weights.keys():
                        raise RuntimeError(
                            'Check the parameter names in ppenalty_weights and ppenalty_dict again. They must match.')
                    else:
                        for k in model.P.keys():
                            if k in ppenalty_dict.keys():
                                terms['L2_penalty'] += ppenalty_weights[k] * (model.P[k] - ppenalty_dict[k]) ** 2

            terms['huplc'] = 0
            obj_weight['huplc'] = 1
            terms['constraint_penalty'] = 0
            obj_weight['constraint_penalty'] = 1
            
            # for new huplc structure (CS):
            if hasattr(model, 'huplc_absorbing'):
                terms['huplc'] = self._huplc_obj_term(model, sigma_sq)
                
                # This might break - I don't have anything to compare it too
                if penaltyparamcon == True: #added for optional penalty term related to constraint CS
                    rho=1e-1
                    sumpen=0.0
                    for t in model.alltime:
                        sumpen = sumpen + m.Y[t,'npen']
                    terms['constraint_penalty'] = rho*sumpen

            expr = 0
            print('expr =')
            for k in terms.keys():
                print(f' + {obj_weight[k]} * {k}')
                expr += terms[k] * obj_weight[k]

            return expr

        # Estimation without model variance and only device variance
        def rule_objective_device_only(model):
            
            terms = {}
            obj_weight = {}
            
            terms['D_bar'] = 0
            obj_weight['D_bar'] = 1
            terms['huplc'] = 0
            obj_weight['huplc'] = 1
            
            if hasattr(model, 'huplc_absorbing'):
                component_set = model._abs_components
            else:
                component_set = list_components
            
            terms['D_bar'] = self._get_objective_expr(model, with_d_vars, component_set, sigma_sq, device=True)
            
            if hasattr(model, 'huplc_absorbing'):
                terms['huplc'] = self._huplc_obj_term(model, sigma_sq)
                      
            expr = 0
            print('expr =')
            for k in terms.keys():
                print(f' + {obj_weight[k]} * {k}')
                expr += terms[k] * obj_weight[k]

            return expr

        if self.model_variance == True:
            model.objective = Objective(rule=rule_objective)
        else:
            model.objective = Objective(rule=rule_objective_device_only)

        custom_weight = 1
        if hasattr(model, 'custom_obj'):
            model.objective.expr += custom_weight*(model.custom_obj)

        if warmstart==True:
            if hasattr(model,'dual') and hasattr(model,'ipopt_zL_out') and hasattr(model,'ipopt_zU_out') and hasattr(model,'ipopt_zL_in') and hasattr(model,'ipopt_zU_in'):
                self.update_warm_start(model)
            else:
                self.add_warm_start_suffixes(model, use_k_aug=False)

        if self.solver in ['k_aug', 'ipopt_sens']:

            if self.solver == 'ipopt_sens':
                hessian = self._covariance_ipopt_sens(model, optimizer, tee, all_sigma_specified)
                
            elif self.solver == 'k_aug':
                hessian = self._covariance_k_aug(model, optimizer, tee, all_sigma_specified)
                
            if self.model_variance:
                hessian_out = self._compute_covariance(hessian, sigma_sq)
                
            print(hessian_out)
                
            self.cov = pd.DataFrame(hessian_out, index=self.param_names, columns=self.param_names)   
            cov_mat = self.cov.values
        
        else:
            solver_results = optimizer.solve(model, tee=tee) 
           
        if with_d_vars:
            model.del_component('D_bar')
            model.del_component('D_bar_constraint')
        
        obj_val = model.objective.expr()
        model.del_component('objective')
        
        return obj_val, cov_mat
        
    
    def _solve_model_given_c(self, sigma_sq, optimizer, **kwds):
        """Solves estimation based on concentration data. (known variances)

        This method is not intended to be used by users directly

        :param dict sigma_sq: variances
        :param SolverFactory optimizer: Pyomo Solver factory object

        :Keyword Args:

            - tee (bool,optional): flag to tell the optimizer whether to stream output
              to the terminal or not

        :return: None

        """
        tee = kwds.pop('tee', False)
        
        if self._huplc_given:
            weights = kwds.pop('weights', [0.0, 1.0, 1.0])
        else:
            weights = kwds.pop('weights', [0.0, 1.0])
       
        covariance = kwds.pop('covariance', False)
        warmstart = kwds.pop('warmstart', False)
        penaltyparamcon = kwds.pop('penaltyparamcon', False) #added for optional penalty term related to constraint CS
        species_list = kwds.pop('subset_components', None)
        cov_mat = None

        list_components = self._get_list_components(species_list)
        model = self.model

        if not self._concentration_given: # and not self._custom_data_given:
            raise NotImplementedError(
                "Parameter Estimation from concentration data requires concentration data model.C[ti,cj]")

        all_sigma_specified = True

        keys = sigma_sq.keys()
        
        # HUPLC shows up here too
        for k in list_components:
            if hasattr(self, 'huplc_absorbing') and k not in keys:
                for ki in self.solid_spec_args1:
                    for key in keys:
                        if key == ki:
                            sigma_sq[ki] = sigma_sq[key]

            elif hasattr(self, 'huplc_absorbing') == False and k not in keys:
                all_sigma_specified = False
                sigma_sq[k] = max(sigma_sq.values())

        def rule_objective(model):
            obj=0
            if penaltyparamcon == True:
                rho = 100
                sumpen = 0.0
                obj += conc_objective(model, variance=sigma_sq)    
                obj += comp_objective(model)
                
                # Is this correct?
                for t in model.allmeas_times:
                    sumpen += model.Y[t, 'npen']
                fifth_term = rho * sumpen
                obj += fifth_term
            else:
                obj += conc_objective(model, variance=sigma_sq)
                obj += comp_objective(model)
            return obj

        model.objective = Objective(rule=rule_objective)

        
        # Add custom objective
        if hasattr(model, 'custom_obj'):
            model.objective.expr += model.custom_obj

        if warmstart==True:
            if hasattr(model,'dual') and hasattr(model,'ipopt_zL_out') and hasattr(model,'ipopt_zU_out') and hasattr(model,'ipopt_zL_in') and hasattr(model,'ipopt_zU_in'):
                self.update_warm_start(model)
            else:
                self.add_warm_start_suffixes(model)


        self.cov = None
        if self.solver in ['k_aug', 'ipopt_sens']:
            
            if self.solver == 'ipopt_sens':
                hessian = self._covariance_ipopt_sens(model, optimizer, tee, all_sigma_specified)
                
            elif self.solver == 'k_aug':
                hessian = self._covariance_k_aug(model, optimizer, tee, all_sigma_specified, labels=True)
                
            self.cov = pd.DataFrame(hessian, index=self.param_names, columns=self.param_names)   
               
        else:
            solver_results = optimizer.solve(model, tee=tee, symbolic_solver_labels=True)
            self._termination_problems(solver_results, optimizer)
            
        obj_val = model.objective.expr()
        model.del_component('objective')
        
        return obj_val, cov_mat
    
    def _covariance_ipopt_sens(self, model, optimizer, tee, all_sigma_specified):
        """Generalize the covariance optimization with IPOPT Sens

        :param ConcreteModel model: The Pyomo model used in parameter fitting
        :param SolverFactory optimizer: The SolverFactory currently being used
        :param bool tee: Display option
        :param dict all_sigma_specified: The provided variances

        :return hessian: The covariance matrix
        :rtype: numpy.ndarray

        """
        if self.model_variance == False:
            print("WARNING: FOR PROBLEMS WITH NO MODEL VARIANCE it is advised to use k_aug!!!")
        self._tmpfile = "ipopt_hess"
        
        solver_results = optimizer.solve(model, 
                                         tee=tee,
                                         logfile=self._tmpfile,
                                         report_timing=True)
        
        print("Done solving building reduce hessian")
        output_string = ''
        with open(self._tmpfile, 'r') as f:
            output_string = f.read()
        if os.path.exists(self._tmpfile):
            os.remove(self._tmpfile)

        ipopt_output, hessian_output = split_sipopt_string(output_string)
        
        if not all_sigma_specified:
            raise RuntimeError(
                'All variances must be specified to determine covariance matrix.\n Please pass variance dictionary to run_opt')

        n_vars = len(self._idx_to_variable)
        print(n_vars)
        
        hessian = read_reduce_hessian(hessian_output, n_vars)
        
        print(f'THE SIZE: {hessian.shape}')
        
        import pandas as pd
        
        self.hessian = pd.DataFrame(hessian)
        
        print(self.hessian)
        
        return hessian
    
    def _covariance_k_aug(self, model, optimizer, tee, all_sigma_specified, labels=False):
        """Generalize the covariance optimization with IPOPT Sens

        :param ConcreteModel model: The Pyomo model used in parameter fitting
        :param SolverFactory optimizer: The SolverFactory currently being used
        :param bool tee: Display option
        :param dict all_sigma_specified: The provided variances
        :param bool labels: Display options

        :return hessian: The covariance matrix
        :rtype: numpy.ndarray

        """
        self.add_warm_start_suffixes(model, use_k_aug=True)   
    
        self._tmpfile = "k_aug_hess"
        ip = SolverFactory('ipopt')
        
        solver_results = ip.solve(model,
                                  tee=tee,
                                  logfile=self._tmpfile,
                                  report_timing=True,
                                  symbolic_solver_labels=True,
                                  keepfiles=True
                                  )
    
        if labels:
            self._termination_problems(solver_results, optimizer)
        
        k_aug = SolverFactory('k_aug')
        self.update_warm_start(model)
        k_aug.options["print_kkt"] = ""
        k_aug.solve(model, tee=True)
        
        inverse_reduced_hessian = self.calculate_inverse_Hr(self._tmpfile, self._idx_to_variable, self.param_names_full)

        return inverse_reduced_hessian
    
    
    @staticmethod
    def calculate_inverse_Hr(tmp_file, variable_index, param_list):
            
        #%%
        # param_names = []
        # tmp_file = r1.p_estimator._tmpfile
        # variable_index = r1.p_estimator._idx_to_variable
        
        # param_names = [p for p in getattr(r1.p_estimator.model, 'P')]
        # param_list = [f'P[{p}]' for p in getattr(r1.p_estimator.model, 'P')]
        # if hasattr(r1.p_estimator.model, 'Pinit'):
        #     param_names += [p for p in getattr(r1.p_estimator.model, 'Pinit')]
        #     param_list += [f'init_conditions[{p}]' for p in getattr(r1.p_estimator.model, 'Pinit')]
        
        
        from pathlib import Path
        import pandas as pd
        import numpy as np
        from scipy.sparse import coo_matrix, csc_matrix, triu
        from scipy.sparse.linalg import spsolve
        from kipet.calculation_tools.reduced_hessian import SparseRowIndexer, delete_from_csr
        
        kaug_files = Path('GJH')

        with open(tmp_file, 'r') as f:
            output_string = f.read()

        stub = output_string.split('\n')[0].split(',')[1][2:-4]

        var_index = pd.read_csv(Path(stub + '.col'), sep=';', header=None)  # dummy sep
        con_index = pd.read_csv(Path(stub + '.row'), sep=';', header=None)  # dummy sep
        var_index_names = [var_name for var_name in var_index[0]]
        con_index_names = [con_name for con_name in con_index[0].iloc[:-1]]
        
        hess_file = kaug_files.joinpath('H_print.txt')
        hess = pd.read_csv(hess_file, delim_whitespace=True, header=None, skipinitialspace=True)
        hess.columns = ['irow', 'jcol', 'vals']
        hess.irow -= 1
        hess.jcol -= 1

        jac_file = kaug_files.joinpath('A_print.txt')
        jac = pd.read_csv(jac_file, delim_whitespace=True, header=None, skipinitialspace=True)
        jac.columns = ['irow', 'jcol', 'vals']
        jac.irow -= 1
        jac.jcol -= 1

        n = len(var_index_names)
        m = len(con_index_names)
        
        J = coo_matrix((jac.vals, (jac.jcol, jac.irow)), shape=(m, n))
        J_c = J.tocsc()
        Hess_coo = coo_matrix((hess.vals, (hess.irow, hess.jcol)), shape=(n, n))
        H = Hess_coo + triu(Hess_coo, 1).T
        
        # H_df = pd.DataFrame(H.todense(), columns=var_index_names, index=var_index_names)
     
        # print(var_index_names)
        # print(n)
        # print(H_df)
     
        col_ind = [var_index_names.index(v.getname()) for v in variable_index.values()]
        # col_dict = {v.getname(): var_index_names.index(v.getname()) for v in variable_index.values()}
        
        col_ind_P = [var_index_names.index(name) for name in param_list]
        col_ind_P_in_Hr = [col_ind.index(p) for p in col_ind_P]
        
        # col_dict = {v.getname(): var_index_names.index(v.getname()) for v in param_list}
        
        row_indexer = SparseRowIndexer(J_c.T)
        J_f = row_indexer[col_ind].T
        J_l = delete_from_csr(J_c.tocsr(), col_indices=col_ind)

        n_free = n - J_f.shape[0]
        
        X = spsolve(J_l.tocsc(), -J_f.tocsc())

        col_ind_left = list(set(range(n)).difference(set(col_ind)))
        col_ind_left.sort()

        Z = np.zeros([n, n_free])
        Z[col_ind, :] = np.eye(n_free)

        if isinstance(X, csc_matrix):
            Z[col_ind_left, :] = X.todense()
        else:
            Z[col_ind_left, :] = X.reshape(-1, 1)

        Z_mat = coo_matrix(np.mat(Z)).tocsr()
        Z_mat_T = coo_matrix(np.mat(Z).T).tocsr()
        Hess = H.tocsr()
        reduced_hessian = Z_mat_T * Hess * Z_mat
        inv_H_r = np.linalg.inv(reduced_hessian.todense())
        
        H_use = inv_H_r[col_ind_P_in_Hr, :]
    
        #%%
    
        return H_use
    
    def _termination_problems(self, solver_results, optimizer):
        """Checks the termination conditions and will try once again if it fails

        :param ResultsObject solver_results: The results from the parameter fittings
        :param SolverFactory optimizer: The solver factory being used

        :return: None

        """
        self.termination_condition = solver_results.solver.termination_condition
        if self.termination_condition != TerminationCondition.optimal:
            print("WARNING: The solution of the iteration was unsuccessful. The problem is solved with additional solver options.")
            optimizer.options["OF_start_with_resto"] = 'yes'
            solver_results = optimizer.solve(self.model, tee=False, symbolic_solver_labels=True)
            self.termination_condition = solver_results.solver.termination_condition
            if self.termination_condition != TerminationCondition.optimal:
                print(
                    "WARNING: The solution of the iteration was unsuccessful. The problem is solved with additional solver options.")
                optimizer.options["OF_start_with_resto"] = 'no'
                # optimizer.options["OF_bound_push"] = 1E-02
                optimizer.options["OF_bound_relax_factor"] = 1E-05
                solver_results = optimizer.solve(self.model, tee=False, symbolic_solver_labels=True)
                self.termination_condition = solver_results.solver.termination_condition
                # options["OF_bound_relax_factor"] = 1E-08
                if self.termination_condition != TerminationCondition.optimal:
                        print("The current iteration was unsuccessful.")
                        
        return None

     #For addition of huplc data and matching liquid and solid species (CS):
        
    def _huplc_obj_term(self, m, sigma_sq):
        """Adds the HUPLC term to the objective

        :param ConcreteModel m: The model used in parameter fitting
        :param dict sigma_sq: The dict of variances

        :return expressions third_term: The HUPLC objective expression

        """
        #print(m, sigma_sq)
        
        
        def rule_Dhat_bar(m, t, l):
            list_huplcabs = [k for k in m.huplc_absorbing.value]
            return m.Dhat_bar[t, l] == (m.Z[t, l] + m.solidvol[t, l]) / (
                sum(m.Z[t, j] + m.solidvol[t, j] for j in list_huplcabs))

        third_term = 0.0
        
        # Arrange the HUPLC device variance
        if not 'device-huplc' in sigma_sq.keys():
            sigma_sq['device-huplc'] = 1.0
            
        if hasattr(m, 'solid_spec_arg1') and hasattr(m, 'solid_spec_arg2'):
            solidvol_dict = dict()
            m.add_component('cons_solidvol', ConstraintList())
            new_consolidvol = getattr(m, 'cons_solidvol')
            m.add_component('solidvol', Var(m.huplctime, m.huplc_absorbing.value, initialize=solidvol_dict))

            for k in m.solid_spec_arg1:
                for j in m.algebraics:
                    for time in m.huplctime:
                        if j == k:
                            for l in m.huplc_absorbing.value:
                                strsolidspec = "\'" + str(k) + "\'" + ', ' + "\'" + str(
                                    l) + "\'"  # pair of absorbing solid and liquid
                                if l in m.solid_spec_arg2 and strsolidspec in str(
                                        m.solid_spec.keys()):  #check whether pair of solid and liquid in keys and whether liquid in huplcabs species
                                    valY = value(m.Y[time, k]) / value(m.vol)
                                    if valY <= 0:
                                        new_consolidvol.add(m.solidvol[time, l] == 0.0)
                                    else:
                                        new_consolidvol.add(m.solidvol[time, l] == m.Y[time, k] / m.vol)
                                else:
                                    new_consolidvol.add(m.solidvol[time, l] == 0.0)
                for jk in self._get_list_components(species_list):
                    for time in m.huplctime:
                        if jk == k:
                            for l in m.huplc_absorbing.value:
                                strsolidspec = "\'" + str(k) + "\'" + ', ' + "\'" + str(
                                    l) + "\'"  # pair of absorbing solid and liquid
                                if l in m.solid_spec_arg2 and strsolidspec in str(
                                        m.solid_spec.keys()):  #check whether pair of solid and liquid in keys and whether liquid in huplcabs species
                                    valZ = value(m.Z[time, k]) / value(m.vol)
                                    if valZ <= 0:
                                        new_consolidvol.add(m.solidvol[time, l] == 0.0)
                                    else:
                                        new_consolidvol.add(m.solidvol[time, l] == m.Z[time, k] / m.vol)
                                else:
                                    new_consolidvol.add(m.solidvol[time, l] == 0.0)

            m.Dhat_bar = Var(m.huplcmeas_times,
                             m.huplc_absorbing)

            m.Dhat_bar_constraint = Constraint(m.huplcmeas_times,
                                               m.huplc_absorbing,
                                               rule=rule_Dhat_bar)

            for t in m.huplcmeas_times:
                list_huplcabs = [k for k in m.huplc_absorbing.value]
                for k in list_huplcabs:
                    third_term += (m.Dhat[t, k] - m.Dhat_bar[t, k]) ** 2 / sigma_sq['device-huplc']

        return third_term

    def _define_reduce_hess_order(self):
        """
        This sets up the suffixes of the reduced hessian

        :return: None

        """
        self.model.red_hessian = Suffix(direction=Suffix.IMPORT_EXPORT)
        count_vars = 1

        if self._spectra_given:
            if self.model_variance:
                count_vars = self._set_up_reduced_hessian(self.model, self.model.times_spectral, self.component_set, self.component_var, count_vars)
                count_vars = self._set_up_reduced_hessian(self.model, self.model.meas_lambdas, self.component_set, 'S', count_vars)
                
        for v in self.model.P.values():
            if v.is_fixed():
                print(v, end='\t')
                print("is fixed")
                continue
            self._idx_to_variable[count_vars] = v
            self.model.red_hessian[v] = count_vars
            count_vars += 1
            
        if hasattr(self.model, 'Pinit'):
            for k, v in self.model.Pinit.items():
                v = self.model.init_conditions[k]
                self._idx_to_variable[count_vars] = v
                self.model.red_hessian[v] = count_vars
                count_vars += 1
               
        return None
        
    def _compute_covariance(self, H, variances):
        """Computes the covariance for post calculation anaylsis

        :param numpy.ndarray hessian: The Hessian matrix
        :param dict variances: Component variances

        :return: The parameter covariances
        
        """
        B = self._compute_B_matrix(variances)
        Vd = self._compute_Vd_matrix(variances)
        
        print(f'B: {B.shape}')
        print(f'H: {H.shape}')
        print(f'Vd: {Vd.shape}')
        
        R = B.T @ H.T
        A = Vd @ R
        L = H @ B
        V_theta = (A.T @ L.T).T

        covariances_p = V_theta
        
        print('V_theta')
        print(V_theta)

        return covariances_p

    def _compute_B_matrix(self, variances, **kwds):
        """Builds B matrix for calculation of covariances

        This method is not intended to be used by users directly

        :param dict variances: variances

        :return: None

        """
        nt = len(self.model.times_spectral)
        time_set = self.model.times_spectral
        
        conc_data_var = 'C'
        nw = len(self.model.meas_lambdas)
        nparams = self._get_nparams(self.model)
       
        if hasattr(self, '_abs_components'):
            n_val = self._nabs_components
            component_set = self._abs_componets
        else:
            n_val = self._n_actual
            component_set = self._sublist_components
    
        variance = variances['device']
        ntheta = n_val * (nw + nt) + nparams
        B_matrix = np.zeros((ntheta, nw * nt))

        for i, t in enumerate(time_set):
            for j, l in enumerate(self.model.meas_lambdas):
                for k, c in enumerate(component_set):
                    r_idx1 = i * n_val + k
                    r_idx2 = j * n_val + k + n_val * nt
                    c_idx = i * nw + j
                    B_matrix[r_idx1, c_idx] = -2 * self.model.S[l, c].value / variance
                    B_matrix[r_idx2, c_idx] = -2 * getattr(self.model, conc_data_var)[t, c].value / variance
                    
        return B_matrix

    def _compute_Vd_matrix(self, variances, **kwds):
        """Builds d covariance matrix

        This method is not intended to be used by users directly

        :param dict variances: variances

        :return: None

        """
        nt = len(self.model.times_spectral)
        nw = len(self.model.meas_lambdas)
        row = []
        col = []
        data = []
        nd = nt * nw
        v_device = variances['device']

        s_array = np.zeros(nw * self.n_val)
        v_array = np.zeros(self.n_val)
        
        for k, c in enumerate(self.component_set):
            v_array[k] = variances[c]

        for j, l in enumerate(self.model.meas_lambdas):
            for k, c in enumerate(self.component_set):
                s_array[j * self.n_val + k] = self.model.S[l, c].value

        for i in range(nt):
            for j in range(nw):
                val = sum(v_array[k] * s_array[j * self.n_val + k] ** 2 for k in range(self.n_val)) + v_device
                row.append(i * nw + j)
                col.append(i * nw + j)
                data.append(val)

                for p in range(nw):
                    if j != p:
                        val = sum(v_array[k] * s_array[j * self.n_val + k] * s_array[p * self.n_val + k] for k in range(self.n_val))
                        row.append(i * nw + j)
                        col.append(i * nw + p)
                        data.append(val)

        Vd_matrix = coo_matrix((data, (row, col)), shape=(nd, nd)).tocsr()
        
        return Vd_matrix

    def run_param_est_with_subset_lambdas(self, builder_clone, end_time, subset, nfe, ncp, sigmas, solver='ipopt'):
        """ Performs the parameter estimation with a specific subset of wavelengths.
            At the moment, this is performed as a totally new Pyomo model, based on the
            original estimation. Initialization strategies for this will be needed.

        :param TemplateBuilder builder_clone: Template builder class of complete model without the data added yet
        :param float end_time: the end time for the data and simulation
        :param list subset: list of selected wavelengths
        :param int nfe: number of finite elements
        :param int ncp: number of collocation points
        :param dict sigmas: dictionary containing the variances, as used in the ParameterEstimator class

        :return ResultsObject results: The solved pyomo model results

        """
        if not isinstance(subset, (list, dict)):
            raise RuntimeError("subset must be of type list or dict!")

        if isinstance(subset, dict):
            lists1 = sorted(subset.items())
            x1, y1 = zip(*lists1)
            subset = list(x1)
        
        # This is the filter for creating the new data subset
        old_D = convert(self.model.D)
        new_D = old_D.loc[:, subset] 
        
        
        print(end_time, new_D)
        # Now that we have a new DataFrame, we need to build the entire problem from this
        # An entire new ParameterEstimation problem should be set up, on the outside of
        # this function and class structure, from the model already developed by the user.
        new_template = construct_model_from_reduced_set(builder_clone, end_time, new_D)
        # need to put in an optional running of the variance estimator for the new
        # parameter estiamtion run, or just use the previous full model run to initialize...

        results, lof = run_param_est(new_template, nfe, ncp, sigmas, solver=solver)

        return results

    def run_lof_analysis(self, builder_before_data, end_time, correlations, lof_full_model, nfe, ncp, sigmas,
                         step_size=0.2, search_range=(0, 1)):
        """ Runs the lack of fit minimization problem used in the Michael's Reaction paper
        from Chen et al. (submitted). To use this function, the full parameter estimation
        problem should be solved first and the correlations for wavelngths from this optimization
        need to be supplied to the function as an option.


        :param TemplateBuilder builder_before_data: Template builder class of complete model without the data added yet
        :param int end_time: the end time for the data and simulation
        :param dict correlations: dictionary containing the wavelengths and their correlations to the concentration profiles
        :param int lof_full_model: the value of the lack of fit of the full model (with all wavelengths)
        :param int nfe: number of finite elements
        :param int ncp: number of collocation points
        :param dict sigmas: dictionary containing the variances, as used in the ParameterEstimator class
        :param float step_size: The spacing used in correlation thresholds
        :param tuple search_range: correlation bounds within to search

        :return: None

        """
        if not isinstance(step_size, float):
            raise RuntimeError("step_size must be a float between 0 and 1")
        elif step_size >= 1 or step_size <= 0:
            return RuntimeError("step_size must be a float between 0 and 1")

        if not isinstance(search_range, tuple):
            raise RuntimeError("search range must be a tuple")
        elif search_range[0] < 0 or search_range[0] > 1 and not (
                isinstance(search_range, float) or isinstance(search_range, int)):
            raise RuntimeError("search range lower value must be between 0 and 1 and must be type float")
        elif search_range[1] < 0 or search_range[1] > 1 and not (
                isinstance(search_range, float) or isinstance(search_range, int)):
            raise RuntimeError("search range upper value must be between 0 and 1 and must be type float")
        elif search_range[1] <= search_range[0]:
            raise RuntimeError("search_range[1] must be bigger than search_range[0]!")
        # firstly we will run the initial search from at increments of 20 % for the correlations
        # we already have lof(0) so we want 10,30,50,70, 90.
        
        count = 0
        filt = 0.0
        initial_solutions = list()
        initial_solutions.append((0, lof_full_model))
        while filt < search_range[1]:
            filt += step_size
            if filt > search_range[1]:
                break
            elif filt == 1:
                break
            new_subs = wavelength_subset_selection(correlations=correlations, n=filt)
            lists1 = sorted(new_subs.items())
            x1, y1 = zip(*lists1)
            x = list(x1)

            old_D = convert(self.model.D)
            new_D = old_D.loc[:, new_subs]

            # opt_model, nfe, ncp = construct_model_from_reduced_set(builder_before_data, end_time, new_D)
            # Now that we have a new DataFrame, we need to build the entire problem from this
            # An entire new ParameterEstimation problem should be set up, on the outside of
            # this function and class structure, from the model already developed by the user.
            new_template = construct_model_from_reduced_set(builder_before_data, end_time, new_D)
            # need to put in an optional running of the variance estimator for the new
            # parameter estimation run, or just use the previous full model run to initialize...
            results, lof = run_param_est(new_template, nfe, ncp, sigmas)
            initial_solutions.append((filt, lof))

            count += 1

        count = 0
        for x in initial_solutions:
            print(f'When wavelengths of less than {x[0]:0.3f} correlation are removed')
            print(f'The lack of fit is {x[1]:0.6f} %')

    # =============================================================================
    # --------------------------- DIAGNOSTIC TOOLS ------------------------
    # =============================================================================

    def lack_of_fit(self):
        """ Runs basic post-processing lack of fit analysis

        :return float lack of fit: percentage lack of fit

        """
        
        if self._spectra_given:
        
            sum_e = 0
            sum_d = 0
            C = np.zeros((len(self.model.times_spectral), self.n_val))
            S = np.zeros((len(self.model.meas_lambdas), self.n_val))
            
            for c_count, c in enumerate(self.component_set):
                for t_count, t in enumerate(self.model.times_spectral):
                    C[t_count, c_count] = getattr(self.model, self.component_var)[t, c].value
    
            for c_count, c in enumerate(self.component_set):
                for l_count, l in enumerate(self.model.meas_lambdas):
                    S[l_count, c_count] = self.model.S[l, c].value
                 
            D_model = C.dot(S.T)
            
            for t_count, t in enumerate(self.model.times_spectral):
                for l_count, l in enumerate(self.model.meas_lambdas):
                    sum_e += (D_model[t_count, l_count] - self.model.D[t, l]) ** 2
                    sum_d += (self.model.D[t, l]) ** 2
      
            lof = np.sqrt(sum_e/sum_d)*100
            
        elif self._concentration_given:
            
            sum_e = 0
            sum_d = 0
           
            for index, values in self.model.Cm.items():
                sum_e += (self.model.Z[index].value - self.model.Cm[index].value) ** 2
                sum_d += (self.model.Cm[index].value) ** 2

            lof = np.sqrt(sum_e/sum_d)*100
            
        #print(f'The lack of fit is {lof:0.2f}%')
        return lof

    def wavelength_correlation(self):
        """ determines the degree of correlation between the individual wavelengths and
        the and the concentrations.


        :return dict: dictionary of correlations with wavelength

        """
        nt = len(self.model.allmeas_times)

        cov_d_l = dict()
        for c in self._sublist_components:
            for l in self.model.meas_lambdas:
                mean_d = (sum(self.model.D[t, l] for t in self.model.times_spectral) / nt)
                mean_c = (sum(self.model.C[t, c].value for t in self.model.times_spectral) / nt)
                cov_d_l[l, c] = 0
                for t in self.model.times_spectral:
                    cov_d_l[l, c] += (self.model.D[t, l] - mean_d) * (self.model.C[t, c].value - mean_c)

                cov_d_l[l, c] = cov_d_l[l, c] / (nt - 1)

        # calculating the standard devs for dl and ck over time
        s_dl = dict()

        for l in self.model.meas_lambdas:
            s_dl[l] = 0
            mean_d = (sum(self.model.D[t, l] for t in self.model.times_spectral) / nt)
            error = 0
            for t in self.model.times_spectral:
                error += (self.model.D[t, l] - mean_d) ** 2
            s_dl[l] = (error / (nt - 1)) ** 0.5

        s_ck = dict()

        for c in self._sublist_components:
            s_ck[c] = 0
            mean_c = (sum(self.model.C[t, c].value for t in self.model.times_spectral) / nt)
            error = 0
            for t in self.model.times_spectral:
                error += (self.model.C[t, c].value - mean_c) ** 2
            s_ck[c] = (error / (nt - 1)) ** 0.5

        cor_lc = dict()

        for c in self._sublist_components:
            for l in self.model.meas_lambdas:
                cor_lc[l, c] = cov_d_l[l, c] / (s_ck[c] * s_dl[l])

        cor_l = dict()
        for l in self.model.meas_lambdas:
            cor_l[l] = max(cor_lc[l, c] for c in self._sublist_components)

        return cor_l

    def lack_of_fit_huplc(self):
        """ Runs basic post-processing lack of fit analysis

        :return float lack of fit: percentage lack of fit

        """
        nc = len(self.model.huplc_absorbing)
        sum_e = 0
        sum_d = 0
        D_model = np.zeros((len(self.model.huplcmeas_times), nc))
        
        for t_count, t in enumerate(self.model.huplcmeas_times):
            for l_count, l in enumerate(self._list_huplcabs):
                
                if hasattr(self.model, 'solidvol'):
                    D_model[t_count, l_count] = (self.model.Z[t, l].value + self.model.solidvol[t, l].value) / (
                        sum(self.model.Z[t, j].value + self.model.solidvol[t, j].value for j in self._list_huplcabs))
                else:
                    D_model[t_count, l_count] = (self.model.Z[t, l].value) / (
                        sum(self.model.Z[t, j].value for j in self._list_huplcabs))

                sum_e += (D_model[t_count, l_count] - self.model.Dhat[t, l].value) ** 2
                sum_d += (self.model.Dhat[t, l].value) ** 2

        lof = np.sqrt((sum_e/sum_d))*100

        print(f"The lack of fit for the huplc data is {lof:0.2f}%")
        return lof
    
    def g_handling_status_messages(self):
        """Determines the unwanted contribution type and informs the user of this

        :return: None

        """
        if self.G_contribution == 'unwanted_G':
            print("\nType of unwanted contributions not set, so assumed that it is time-variant.\n")
            self.G_contribution = 'time_variant_G'
        elif self.G_contribution == 'time_variant_G':
            print("\nTime-variant unwanted contribution is involved.\n")
        elif self.G_contribution == 'time_invariant_G_decompose':
            print("\nTime-invariant unwanted contribution is involved and G can be decomposed.\n")
        elif self.G_contribution == 'time_invariant_G_no_decompose':
            print("\nTime-invariant unwanted contribution is involved but G cannot be decomposed.\n")
        return None
     
    def decompose_G_test(self, St, Z_in):
        """Check whether or not G can be decomposed

        :param dict St: Reaction coefficient matrix
        :param dict Z_in: Dosing points

        :return: None

        """
        if St == dict() and Z_in == dict():
            raise RuntimeError('Because time-invariant unwanted contribution is chosen, please provide information of St or Z_in to build omega matrix.')
        
        omega_list = [St[i] for i in St.keys()]
        omega_list += [Z_in[i] for i in Z_in.keys()]
        omega_sub = np.array(omega_list)
        rank = np.linalg.matrix_rank(omega_sub)
        cols = omega_sub.shape[1]
        rko = cols - rank
        
        if rko > 0:
            self.time_invariant_G_decompose = True
            self.G_contribution = 'time_invariant_G_decompose'
        else:
            self.time_invariant_G_no_decompose = True
            self.G_contribution = 'time_invariant_G_no_decompose'
        return None

def wavelength_subset_selection(correlations=None, n=None):
    """ identifies the subset of wavelengths that needs to be chosen, based
    on the minimum correlation value set by the user (or from the automated
    lack of fit minimization procedure)

    :param dict correlations: dictionary obtained from the wavelength_correlation  function, containing every
       wavelength from the original set and their correlations to the concentration profile.
    :param int n: a value between 0 - 1 that slects the minimum amount correlation between the wavelength and the
       concentrations.

    :return: A dictionary of correlations with wavelength
    :rtype: dict

    """
    if not isinstance(correlations, dict):
        raise RuntimeError("correlations must be of type dict()! Use wavelength_correlation function first!")

    # should check whether the dictionary contains all wavelengths or not
    if not isinstance(n, float):
        raise RuntimeError("n must be of type int!")
    elif n > 1 or n < 0:
        raise RuntimeError("n must be a number between 0 and 1")

    subset_dict = dict()
    for l in correlations.keys():
        if correlations[l] >= n:
            subset_dict[l] = correlations[l]
    return subset_dict


# =============================================================================
# ----------- PARAMETER ESTIMATION WITH WAVELENGTH SELECTION ------------------
# =============================================================================

def construct_model_from_reduced_set(builder_clone, end_time, D):
    """ constructs the new pyomo model based on the selected wavelengths.

    :param TemplateBuilder builder_clone: Template builder class of complete model without the data added yet
    :param float end_time: The end time for the data and simulation
    :param pandas.DataFrame D: The new, reduced dataset with only the selected wavelengths.

    :return TemplateBuilder opt_mode: new Pyomo model from TemplateBuilder, ready for parameter estimation

    """
    import pandas as pd
    from kipet.model_tools.template_builder import TemplateBuilder
    if not isinstance(builder_clone, TemplateBuilder):
        raise RuntimeError('builder_clone needs to be of type TemplateBuilder')

    if not isinstance(D, pd.DataFrame):
        raise RuntimeError('Spectral data format not supported. Try pandas.DataFrame')

    builder_clone._spectral_data = D
    opt_model = builder_clone.create_pyomo_model(0.0, end_time)

    return opt_model


def run_param_est(opt_model, nfe, ncp, sigmas, solver='ipopt'):
    """ Runs the parameter estimator for the selected subset

    :param ConcreteModel opt_model: The model that we wish to run the
    :param int nfe: The number of finite elements
    :param int ncp: The number of collocation points
    :param dict sigmas: A dictionary containing the variances, as used in the ParameterEstimator class

    :return results_pyomo: Parameter Estimation results
    :return lof: lack of fit results

    """
    p_estimator = ParameterEstimator(opt_model)
    p_estimator.apply_discretization('dae.collocation', nfe=nfe, ncp=ncp, scheme='LAGRANGE-RADAU')
    options = dict()

    

    # These may not always solve, so we need to come up with a decent initialization strategy here
    if solver == 'ipopt':
        results_pyomo = p_estimator.run_opt('ipopt',
                                            tee=False,
                                            solver_opts=options,
                                            variances=sigmas)
    else:
        results_pyomo = p_estimator.run_opt(solver,
                                            tee=False,
                                            solver_opts=options,
                                            variances=sigmas,
                                            covariance=True)
    
    lof = p_estimator.lack_of_fit()

    return results_pyomo, lof



"""
Classes used to manage the data in KIPET
"""
# Third party imports
import pandas as pd
import plotly.graph_objs as go

# KIPET library imports
from kipet.input_output.kipet_io import read_file
from kipet.visuals.plots import colors


class DataBlock:
    """Class to manage data in KIPET.
    
    This class is not intended to be used directly by the user.
    
    :Methods:
        
        - :func:`add_dataset`
        - :func:`names`
        - :func:`time_span`
        
    """

    def __init__(self):
        """Initialize the DataBlock instance
        
        This class is intialized without any attributes.
        
        """
        self.datasets = {}

    def __getitem__(self, value):

        return self.datasets[value]

    def __str__(self):

        format_string = "{:<20}{:<15}{:<30}{:<15}{:<30}\n"
        data_str = 'DataBlock:\n'
        data_str += format_string.format(*['Name', 'Category', 'Components', 'Size', 'File'])

        for dataset in self.datasets.values():

            species_str = ', '.join([s for s in dataset.species[:3]])
            if len(dataset.species) > 3:
                species_str += ', ...'

            if dataset.file is not None:
                data_str += format_string.format(f'{dataset.name}', f'{dataset.category}', f'{species_str}',
                                                 f'{dataset.data.shape}', f'{dataset.file.name}')
            else:
                data_str += format_string.format(f'{dataset.name}', f'{dataset.category}', f'{species_str}',
                                                 f'{dataset.data.shape}', 'None')

        return data_str

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        for param, data in self.datasets.items():
            yield data

    def __len__(self):
        return len(self.datasets)

    def add_dataset(self, dataset):
        """Method to add a dataset to the DataBlock

        This takes in information regarding a dataset, creates the DataSet object, and adds the DataSet to
        the datasets attribute.
        
        :param DataSet dataset: A DataSet object to add to the DataBlock
        
        :return: None
        
        """
        self.datasets[dataset.name] = dataset

        return None

    @property
    def _check_duplicates(self):
        """Checks for duplications of data for a specific model element. This does not check spectral or trajectory
        data sets.

        :Raises:

            A ValueError if a component has more than one data set attributed to it.

        :return: None

        """
        all_data_cols = []
        for dataset in self.datasets.values():
            if dataset.category in ['spectral', 'trajectory']:
                continue
            all_data_cols.extend(dataset.data.columns)

        num_duplicates = len(all_data_cols) - len(set(all_data_cols))
        if num_duplicates != 0:
            raise ValueError('Duplicates ({num_duplicates}) in component, state, and algebraic data detected!')

        return None

    @property
    def names(self):
        """Returns the list of dataset names
        
        :return: List of names
        :rtype: list
        
        """
        return [dataset.name for dataset in self.datasets.values()]

    @property
    def time_span(self):
        """Determine the span of time from the beginning to the end of the 
        measured data
        
        :return: the start and end times
        :rtype: tuple
        
        """
        start_time = min([dataset.time_span[0] for dataset in self.datasets.values()])
        stop_time = max([dataset.time_span[1] for dataset in self.datasets.values()])

        return start_time, stop_time


class DataSet:
    """The specific data object contained in the DataBlock
    
    This class is not intended to be used directly by the user.
    
    :Methods:
        
        - :func:`remove_datasets`
        - :func:`show_data`
        - :func:`shape`
        - :func:`size`
        - :func:`species`
        - :func:`time_span`
        
    """

    def __init__(self,
                 name=None,
                 category=None,
                 data=None,
                 file=None,
                 ):

        self.name = name
        self.category = category
        self.data = data
        self.file = file
        self._check_input()

    def __repr__(self):

        if self.data is not None:
            return f'DataSet(name={self.name}, category={self.category}, data={self.data.shape})'
        else:
            return f'DataSet(name={self.name}, category={self.category}, data=None)'

    def __str__(self):

        if self.data is not None:
            return f'DataSet(name={self.name}, category={self.category}, data={self.data.shape})'
        else:
            return f'DataSet(name={self.name}, category={self.category}, data=None)'

    def _check_input(self):
        """Checks whether the data has been provided as a pandas dataframe or
        if a filename has been provided. If not, the DataBlock is initialized
        without any data.
        
        :return: None
        
        """
        if self.data is not None:
            if not isinstance(self.data, pd.DataFrame):
                raise ValueError('Data must be a pandas DataFrame instance')

        elif self.data is None and self.file is not None:
            print(self.file)
            self.data = read_file(self.file)

        else:
            print('DataBlock object initialized without data')

        return None

    def remove_negatives(self):
        """Replaces the negative values in the dataframe with zeros
        
        :return: None
        
        """
        # self.data[self.data < 0] = 0
        self.data.mask(self.data < 0, 0)
        return None

    def show_data(self):
        """Method to show the data using Plotly
        
        This will first determine if the proper data has been provided for
        plotting and will then call the correct plotting method depending on
        the data category.
        
        :return: None
        
        """
        if self.data is None:
            print('No data to plot')
            return None

        if self.category in ['concentration', 'state', 'trajectory', 'custom']:
            self._plot_2d_data()

        if self.category == 'spectral':
            self._plot_spectral_data()

        return None

    def _plot_2d_data(self):
        """Simple plots for showing concentration or complementary state data
        
        :return: None
        
        """
        y_axis_text = 'Measured Value'
        if self.description is not None:
            y_axis_text = self.description

        if self.units is not None:
            y_axis_text = ' '.join([y_axis_text, '[' + self.units[0] + ']'])

        x_axis_text = 'Time'

        if self.units is not None and len(self.units) == 2:
            x_axis_text = ' '.join([x_axis_text, '[' + self.units[1] + ']'])

        fig = go.Figure()
        for i, cols in enumerate(self.data.columns):
            fig.add_trace(go.Scatter(x=self.data.index,
                                     y=self.data[cols],
                                     mode='markers',
                                     name=cols,
                                     marker=dict(size=10, opacity=0.5, color=colors[i])),
                          )

        fig.update_layout(title_text=f'{self.category.capitalize()} data: {self.name}',
                          xaxis_title=f'{x_axis_text}',
                          yaxis_title=f'{y_axis_text}',
                          title_font_size=30)

        fig.plot()

        return None

    def _plot_spectral_data(self):
        """ Plots spectral data in 3D plot
        
        :return: None
        
        """
        fig = go.Figure()
        fig.add_trace(go.Surface(x=self.data.columns,
                                 y=self.data.index,
                                 z=self.data.values,
                                 ))

        fig.update_layout(scene=dict(
            xaxis_title='Wavelength',
            yaxis_title='Time',
            zaxis_title='Spectra'),
            margin=dict(r=100, b=50, l=100, t=50),
            title_text=f'{self.name}: {self.category.capitalize()} data',
            title_font_size=30)

        fig.plot()

        return None

    # Properties
    @property
    def shape(self):
        """Get the shape of the data
        
        :return: shape data or None
        :rtype: tuple/None
        
        """
        if self.data is not None:
            return self.data.shape
        else:
            return None

    @property
    def size(self):
        """Get the size of the data
        
        :return: length of the data
        :rtype: int/None
        
        """
        if self.data is not None:
            return len(self.data.columns)
        else:
            return None

    @property
    def species(self):
        """Get the names of the elements in the dataset
        
        :return: Names of the data elements
        :rtype: list
        
        """
        if self.data is not None and self.category in ['concentration', 'state', 'trajectory', 'custom']:
            return list(self.data.columns)
        elif self.data is None:
            return []
        else:
            return []

    @property
    def time_span(self):
        """Get the time span of the measured data in the dataset
        
        :return: The start and end times of the measured data
        :rtype: tuple
        
        """
        return self.data.index.min(), self.data.index.max()


if __name__ == '__main__':
    pass



"""
Read/Write functions used for data in Kipet
"""
import sys
from contextlib import contextmanager

@contextmanager
def stdout_redirector(stream):
    old_stdout = sys.stdout
    sys.stdout = stream
    try:
        yield
    finally:
        sys.stdout = old_stdout



"""
Top Level helper classes for KIPET
"""


class AttrDict(dict):

    """This class lets you use nested dicts like accessing attributes using a dot notation

    :Methods:

        - :func:`update`

    """
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)
    
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setitem__(self, key, item):
        
        if isinstance(item, dict):
            return super(AttrDict, self).__setitem__(key, AttrDict(item))
        else:
            return dict.__setitem__(self, key, item)

    def update(self, *args, **kwargs):
        """Method used to update the AttrDict using the same syntax as the dict class

        :param tuple args: The arguments (pass through)
        :param dict kwargs: The keyword arguments (pass through)

        :return: None
        """
        for k, v in dict(*args, **kwargs).items():
            self[k] = v

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


class DosingPoint:
    """Small class to handle the dosing points in a cleaner manner

    :param str component: The name of the model component
    :param float time: The time of the dosing
    :param tuple conc: A tuple of the concentration (conc (float), units (str))
    :param tuple vol: A tuple of the dosing volume (vol (float), units (str))

    :Methods:

        - :func:`as_list`

    """
    def __init__(self, component, time, conc, vol):
        
        self.component = component
        self.time = time
        self.conc = conc
        self.vol = vol
    
    def __repr__(self):
        return f'{self.component}: {self.time}, {self.conc} {self.vol}'
    
    def __str__(self):
        return self.__repr__()
    
    @property
    def as_list(self):
        """Return the dosing attributes as a list

        :return: List of dosing point attributes
        :rtype: list

        """
        return [self.component, self.time, self.conc, self.vol]



"""
Parameter handling for ConcreteModels in various cases (especially MEE)
"""
# KIPET library imports
from kipet.general_settings.variable_names import VariableNames

# Load the KIPET variables
__var = VariableNames()


def calculate_parameter_averages(model_dict):
    """Calculate the average parameter values across experiments.

    This takes a plural value of ReactionModel ResultsObject (they need to be already fit) and determines the average
    parameter values. This is meant to be used as a possible initialization in MEE (not necessary).

    :param dict model_dict: The dict of ReactionModel instances

    :return avg_param: A dict of parameter averages
    :rtype: dict

    """
    
    p_dict = {}
    c_dict = {}
    
    for key, model in model_dict.items():
        for param, obj in getattr(model, __var.model_parameter).items():
            if param not in p_dict:
                p_dict[param] = obj.value
                c_dict[param] = 1
            else:
                p_dict[param] += obj.value
                c_dict[param] += 1
                
    avg_param = {param: p_dict[param]/c_dict[param] for param in p_dict.keys()}
    
    return avg_param


def initialize_parameters(model_dict):
    """Automates the parameter initialization in MEE by first calculation parameter averages. This changes the values
    in place.

    :param dict model_dict: The dict of ReactionModel instances

    :return: None

    """
    avg_param = calculate_parameter_averages(model_dict)
    
    for key, model in model_dict.items():
        for param, obj in getattr(model, __var.model_parameter).items():
            obj.value = avg_param[param] 
            
    return None


def check_initial_parameter_values(model_object):
    """Checks the initial parameter values and alters them if they violate the
    bounds.
    
    :param ConcreteModel model_object: A pyomo model instance of the current problem
            
    :return: None
        
    """
    for k, v in getattr(model_object, __var.model_parameter).items():
        
        bound_push = 1.05
        
        if v.value >= v.ub:
            v.set_value(param.ub/bound_push)
        
        if v.value <= v.lb:
            v.set_value(v.lb*bound_push)
            
    return None


def set_scaled_parameter_bounds(model_object, parameter_set=None, rho=10, scaled=True, original_bounds=None):
    """Set the parameter values (scaled) for a given set of parameters. This is done in place.

    :param ConcreteModel model_object: A pyomo model instance of the current problem
            
    :param list parameter_set: A list of parameters to be considered, if None all parameters in the model are considered
    :param float rho: Ratio for setting the upper and lower bounds
    :param bool scaled: True if the parameters are scaled
    :param bool original_bounds: True if the original parameter bounds are to be used (instead of rho)
            
    :return: None
    
    """
    param_set = param = getattr(model_object, __var.model_parameter)
    
    if parameter_set is None:
        parameter_set = [p for p in param_set]
    for k, v in param_set.items():
        if k in parameter_set:
            if not scaled:
                ub = rho*v.value
                lb = 1/rho*v.value
            else:
                ub = rho
                lb = 1/rho
            if original_bounds is not None:
                if ub > original_bounds[k][1]:
                    ub = original_bounds[k][1]
                if lb < original_bounds[k][0]:
                    lb = original_bounds[k][0]
                
            v.setlb(lb)
            v.setub(ub)
            v.unfix()
            if scaled:
                v.set_value(1)
        else:
            v.fix()

    return None



"""
Module to hold the method developed by Chen et al. 2016
"""
from pyomo.environ import Objective, SolverFactory


def solve_C(var_est_object, solver, **kwds):
    """Solves formulation 23 in Chen. et al 2016 for C

    :param VarianceEstimator var_est_object: The variance estimation object
    :param str solver: The solver being used (currently not used)
    :param dict kwds: The dict of user settings passed on from the ReactionModel

    :return: None

    """
    solver_opts = kwds.pop('solver_opts', dict())
    tee = kwds.pop('tee', False)
    update_nl = kwds.pop('update_nl', False)
    profile_time = kwds.pop('profile_time', False)

    # D_bar obj
    obj = 0.0
    for t in var_est_object.model.times_spectral:
        for l in var_est_object._meas_lambdas:
            D_bar = sum(var_est_object.model.S[l, k].value*var_est_object.C_model.C[t, k] for k in var_est_object.comps['absorbing'])
            obj += (var_est_object.model.D[t, l]-D_bar)**2
     
    var_est_object.C_model.objective = Objective(expr=obj)
             
    if profile_time:
        print('-----------------Solve_C--------------------')

    opt = SolverFactory(solver)

    for key, val in solver_opts.items():
        opt.options[key]=val
         
    solver_results = opt.solve(var_est_object.C_model,
                               logfile=var_est_object._tmp4,
                               tee=tee,
                               #keepfiles=True,
                               report_timing=profile_time)

    var_est_object.C_model.del_component('objective')
     
    for t in var_est_object.model.allmeas_times:
        for c in var_est_object.comps['absorbing']:
            var_est_object.model.C[t, c].value = var_est_object.C_model.C[t, c].value

    return None


def solve_S(var_est_object, solver, **kwds):
    """Solves formulation 23 in Chen. et al 2016 for S

    :param VarianceEstimator var_est_object: The variance estimation object
    :param str solver: The solver being used (currently not used)
    :param dict kwds: The dict of user settings passed on from the ReactionModel

    :return: None

    """
    solver_opts = kwds.pop('solver_opts', dict())
    tee = kwds.pop('tee', False)
    update_nl = kwds.pop('update_nl', False)
    profile_time = kwds.pop('profile_time', False)
        
    for l in var_est_object._meas_lambdas:
        for c in var_est_object.comps['unknown_absorbance']:
            var_est_object.S_model.S[l, c].value = var_est_object.model.S[l, c].value
            if hasattr(var_est_object.model, 'known_absorbance'):
                if c in var_est_object.model.known_absorbance:
                    if var_est_object.model.S[l, c].value != var_est_object.model.known_absorbance_data[c][l]:
                        var_est_object.model.S[l, c].set_value(var_est_object.model.known_absorbance_data[c][l])
                        var_est_object.S_model.S[l, c].fix()
    
    #D_bar obj
    obj = 0.0
    for t in var_est_object.model.times_spectral:
        for l in var_est_object._meas_lambdas:
            D_bar = sum(var_est_object.S_model.S[l, k] * var_est_object.model.Z[t, k].value for k in var_est_object.comps['unknown_absorbance'])
            obj += (D_bar - var_est_object.model.D[t, l]) ** 2
                
    var_est_object.S_model.objective = Objective(expr=obj)

    if profile_time:
        print('-----------------Solve_S--------------------')

    opt = SolverFactory(solver)

    for key, val in solver_opts.items():
        opt.options[key]=val
    solver_results = opt.solve(var_est_object.S_model,
                               logfile=var_est_object._tmp3,
                               tee=tee,
                               #keepfiles=True,
                               #show_section_timing=True,
                               report_timing=profile_time)

    var_est_object.S_model.del_component('objective')
    
    for l in var_est_object._meas_lambdas:
        for c in var_est_object.comps['unknown_absorbance']:
            var_est_object.model.S[l, c].value = var_est_object.S_model.S[l, c].value
            if hasattr(var_est_object.model, 'known_absorbance'):
                if c in var_est_object.model.known_absorbance:
                    if var_est_object.model.S[l, c].value != var_est_object.model.known_absorbance_data[c][l]:
                        var_est_object.model.S[l, c].set_value(var_est_object.model.known_absorbance_data[c][l])
                        var_est_object.S_model.S[l, c].fix()
               

def solve_Z(var_est_object, solver, **kwds):
    """Solves formulation 20 in Chen. et al 2016

    :param VarianceEstimator var_est_object: The variance estimation object
    :param str solver: The solver being used (currently not used)
    :param dict kwds: The dict of user settings passed on from the ReactionModel

    :return: None

    :Keyword Args:

        - solver_opts (dict, optional): options passed to the nonlinear solver
        - tee (bool,optional): flag to tell the optimizer whether to stream output
        to the terminal or not
        - profile_time (bool,optional): flag to tell pyomo to time the construction and solution of the model.
        Default False
        - subset_lambdas (array_like,optional): Set of wavelengths to used in initialization problem
        (Weifeng paper). Default all wavelengths.

    """
    solver_opts = kwds.pop('solver_opts', dict())
    tee = kwds.pop('tee', False)
    profile_time = kwds.pop('profile_time', False)
    
    for t in var_est_object.model.allmeas_times:
        for k in var_est_object.comps['unknown_absorbance']:
            if hasattr(var_est_object.model, 'non_absorbing'):
                if k not in var_est_object.model.non_absorbing:
                    pass
                else:
                    var_est_object.model.C[t, k].fixed = True

    obj = 0.0
    
    # Conc obj no sigma - will this work with missing data?
    for k in var_est_object.comps['unknown_absorbance']:
        obj += sum((var_est_object.model.C[t, k]-var_est_object.model.Z[t, k])**2 for t in var_est_object.model.times_spectral)
        

    var_est_object.model.z_objective = Objective(expr=obj)
    if profile_time:
        print('-----------------Solve_Z--------------------')
        
    opt = SolverFactory(solver)

    for key, val in solver_opts.items():
        opt.options[key]=val

    solver_results = opt.solve(var_est_object.model,
                               logfile=var_est_object._tmp2,
                               tee=tee,
                               #show_section_timing=True,
                               report_timing=profile_time)

    var_est_object.model.del_component('z_objective')



"""
Initialization for method from Chen et al. 2016 
"""
# Standard library imports
import os

# Third party imports
import numpy as np
import pandas as pd
from pyomo.environ import Objective, SolverFactory
from scipy.optimize import least_squares

# KIPET library imports
from kipet.calculation_tools.beer_lambert import S_from_DC
from kipet.estimator_tools.results_object import ResultsObject
from kipet.variance_methods.chen_method_pyomo import solve_C, solve_S, solve_Z
from kipet.variance_methods.chen_method_scipy import (build_c_model,
                                                      build_s_model, solve_c_scipy, solve_s_scipy)


def run_method(var_est_object, solver, run_opt_kwargs):
    """This is the original method for estimating variances from Chen et
    al. 2016

    This is an improved method for determining the component variances. This method has been removed from the
    VarianceEstimator class for simplification.

    :param VarianceEstimator var_est_object: The variance estimation object
    :param str solver: The solver being used (currently not used)
    :param dict run_opt_kwargs: The dict of user settings passed on from the ReactionModel

    :return results: The results from the variance estimation
    :rtype: ResultsObject

    """
    solver_opts = run_opt_kwargs.pop('solver_opts', dict())
    max_iter = run_opt_kwargs.pop('max_iter', 400)
    tol = run_opt_kwargs.pop('tolerance', 5.0e-5)
    tee = run_opt_kwargs.pop('tee', False)
    norm_order = run_opt_kwargs.pop('norm', np.inf)
    A = run_opt_kwargs.pop('subset_lambdas', None)
    init_C = run_opt_kwargs.pop('init_C', None)
    lsq_ipopt = run_opt_kwargs.pop('lsq_ipopt', False)
    species_list = run_opt_kwargs.pop('subset_components', None)
    fixed_device_var = run_opt_kwargs.pop('fixed_device_variance', None)

    if init_C is None:
        solve_initalization(var_est_object, 
                            solver,
                            subset_lambdas=A,
                            solver_opts=solver_opts,
                            tee=tee)
    else:
        for t in var_est_object.model.times_spectral:
            for k in var_est_object.comps['unknown_absorbance']:
                var_est_object.model.C[t, k].value = init_C[k][t]
                var_est_object.model.Z[t, k].value = init_C[k][t]

        # This comes from Optimizer
        s_array = S_from_DC(var_est_object.model, init_C)
        S_frame = pd.DataFrame(data=s_array,
                               columns=var_est_object.comps['unknown_absorbance'],
                               index=var_est_object._meas_lambdas)

        if hasattr(var_est_object, '_abs_components'):
            component_set = var_est_object._abs_components
        else:
            component_set = var_est_object._mixture_components

        for l in var_est_object._meas_lambdas:
            for k in component_set:
                var_est_object.model.S[l, k].value = S_frame[k][l]  # 1e-2
                if hasattr(var_est_object.model, 'known_absorbance'):
                    if k in var_est_object.model.known_absorbance:
                        var_est_object.model.S[l, k].value = var_est_object.model.known_absorbance_data[k][l]

    print("{: >11} {: >20}".format('Iter', '|Zi-Zi+1|'))
    logiterfile = "iterations.log"
    if os.path.isfile(logiterfile):
        os.remove(logiterfile)

    if lsq_ipopt:
        build_s_model(var_est_object)
        build_c_model(var_est_object)
    else:
        if species_list is None:
            build_scipy_lsq_arrays(var_est_object)
        else:
            lsq_ipopt = True
            build_s_model(var_est_object)
            build_c_model(var_est_object)

    for it in range(max_iter):

        rb = ResultsObject()

        # vars_to_load = ['Z', 'C', 'Cs', 'S', 'Y']
        # if not hasattr(var_est_object, '_abs_components'):
        #     vars_to_load.remove('Cs')
        rb.load_from_pyomo_model(var_est_object.model)#, to_load=vars_to_load)

        solve_Z(var_est_object, solver)

        if lsq_ipopt:
            solve_S(var_est_object, solver)
            solve_C(var_est_object, solver)
        else:
            solved_s = solve_s_scipy(var_est_object)
            solved_c = solve_c_scipy(var_est_object)

        ra = ResultsObject()
        ra.load_from_pyomo_model(var_est_object.model)#, to_load=vars_to_load)

        r_diff = compute_diff_results(rb, ra)
        Z_norm = r_diff.compute_var_norm('Z', norm_order)

        if it > 0:
            print("{: >11} {: >20}".format(it, Z_norm))
        _log_iterations(var_est_object, logiterfile, it)
        if Z_norm < tol and it >= 1:
            break

    results = ResultsObject()

    results.load_from_pyomo_model(var_est_object.model)

    print('Iterative optimization converged. Estimating variances now')
    solved_variances = _solve_variances(var_est_object,
                                        results,
                                        fixed_dev_var=fixed_device_var)

    compute_D_given_SC(var_est_object, results)

    results.P = {name: var_est_object.model.P[name].value for name in var_est_object.model.parameter_names}

    # removes temporary files. This needs to be changes to work with pyutilib
    if os.path.exists(var_est_object._tmp2):
        os.remove(var_est_object._tmp2)
    if os.path.exists(var_est_object._tmp3):
        os.remove(var_est_object._tmp3)
    if os.path.exists(var_est_object._tmp4):
        os.remove(var_est_object._tmp4)

    return results


def solve_initalization(var_est_object, solver, **kwds):
    """Solves formulation 19 in Chen. et al 2016

    :param VarianceEstimator var_est_object: The variance estimation object
    :param str solver: The solver being used (currently not used)
    :param dict kwds: The dict of user settings passed on from the ReactionModel

    :return: None

    """
    solver_opts = kwds.pop('solver_opts', dict())
    tee = kwds.pop('tee', True)
    set_A = kwds.pop('subset_lambdas', list())
    profile_time = kwds.pop('profile_time', False)
    sigmas_sq = kwds.pop('variances', dict())

    if not set_A:
        set_A = var_est_object.model.meas_lambdas

    keys = sigmas_sq.keys()
    # added due to new structure for non_abs species, non-absorbing species not included in S and Cs as subset of C (CS):

    for k in var_est_object.comps['unknown_absorbance']:
        if k not in keys:
            sigmas_sq[k] = 0.0

    print("Solving Initialization Problem\n")

    # Check this!!!!
    var_est_object._warn_if_D_negative()

    obj = 0.0
    for t in var_est_object.model.times_spectral:
        for l in set_A:
            D_bar = sum(
                var_est_object.model.Z[t, k] * var_est_object.model.S[l, k] for k in var_est_object.comps['unknown_absorbance'])
            obj += (var_est_object.model.D[t, l] - D_bar) ** 2

    var_est_object.model.init_objective = Objective(expr=obj)

    opt = SolverFactory(solver)

    for key, val in solver_opts.items():
        opt.options[key] = val
    solver_results = opt.solve(var_est_object.model,
                               tee=tee,
                               report_timing=profile_time)

    for t in var_est_object.model.times_spectral:
        for k in var_est_object._mixture_components:
            if k in sigmas_sq and sigmas_sq[k] > 0.0:
                var_est_object.model.C[t, k].value = np.random.normal(var_est_object.model.Z[t, k].value, sigmas_sq[k])
            else:
                var_est_object.model.C[t, k].value = var_est_object.model.Z[t, k].value

    var_est_object.model.del_component('init_objective')

    return None


def _log_iterations(var_est_object, filename, iteration):
    """Log solution of each sub-problem in Chen et al. 2016

    :param VarianceEstimator var_est_object: The variance estimation object
    :param str filename: The filename of the tmp files
    :param int iteration: The iteration number

    :return: None

    """
    with open(filename, "a") as f:
        f.write("\n#######################Iteration {}#######################\n".format(iteration))
        with open(var_est_object._tmp2, 'r') as tf:
            f.write("\n#######################Solve Z {}#######################\n".format(iteration))
            f.write(tf.read())
        with open(var_est_object._tmp3, 'r') as tf:
            f.write("\n#######################Solve S {}#######################\n".format(iteration))
            f.write(tf.read())
        with open(var_est_object._tmp4, 'r') as tf:
            f.write("\n#######################Solve C {}#######################\n".format(iteration))
            f.write(tf.read())

    return None


def _solve_variances(var_est_object, results, fixed_dev_var=None):
    """Solves Eq. 23 in Chen et al. 2016

    :param VarianceEstimator var_est_object: The variance estimation object
    :param ResultsObject results: The results object
    :param float fixed_dev_var: Optional device variance

    :return: Indication if variances were estimated successfully.
    :rtype: bool

    """
    nl = len(var_est_object.model.meas_lambdas)
    nt = len(var_est_object.model.times_spectral)
    b = np.zeros((nl, 1))

    variance_dict = dict()
    n_val = len(var_est_object.comps['unknown_absorbance'])

    A = np.ones((nl, n_val + 1))
    reciprocal_nt = 1.0 / nt
    for i, l in enumerate(var_est_object.model.meas_lambdas):
        for j, t in enumerate(var_est_object.model.times_spectral):
            D_bar = 0.0
            for w, k in enumerate(var_est_object.comps['unknown_absorbance']):
                A[i, w] = results.S[k][l] ** 2
                D_bar += results.S[k][l] * results.Z[k][t]
            b[i] += (var_est_object.model.D[t, l] - D_bar) ** 2
        b[i] *= reciprocal_nt

    if fixed_dev_var == None:
        res_lsq = np.linalg.lstsq(A, b, rcond=None)
        all_nonnegative = True
        n_vars = n_val + 1

        for i in range(n_vars):
            if res_lsq[0][i] < 0.0:
                if res_lsq[0][i] < -1e-5:
                    all_nonnegative = False
                else:
                    res_lsq[0][i] = abs(res_lsq[0][i])
            res_lsq[0][i]

        if not all_nonnegative:
            x0 = np.zeros(n_val + 1) + 1e-2
            bb = np.zeros(nl)
            for i in range(nl):
                bb[i] = b[i]

            def F(x, M, rhs):
                return rhs - M.dot(x)

            def JF(x, M, rhs):
                return -M

            res_lsq = least_squares(F, x0, JF,
                                    bounds=(0.0, np.inf),
                                    verbose=2, args=(A, bb))
            for i, k in enumerate(var_est_object.comps['unknown_absorbance']):
                variance_dict[k] = res_lsq.x[i]
            variance_dict['device'] = res_lsq.x[n_val]
            results.sigma_sq = variance_dict
            return res_lsq.success

        else:
            for i, k in enumerate(var_est_object.comps['unknown_absorbance']):
                variance_dict[k] = res_lsq[0][i][0]
            variance_dict['device'] = res_lsq[0][n_val][0]
            results.sigma_sq = variance_dict

    if fixed_dev_var:
        bp = np.zeros((nl, 1))
        Ap = np.zeros((nl, n_val))
        for i, l in enumerate(var_est_object.model.meas_lambdas):
            bp[i] = b[i] - fixed_dev_var
            for j, t in enumerate(var_est_object.model.times_spectral):
                for w, k in enumerate(var_est_object.comps['unknown_absorbance']):
                    Ap[i, w] = results.S[k][l] ** 2

        res_lsq = np.linalg.lstsq(Ap, bp, rcond=None)
        all_nonnegative = True
        n_vars = n_val
        for i in range(n_vars):
            if res_lsq[0][i] < 0.0:
                if res_lsq[0][i] < -1e-5:
                    all_nonnegative = False
                else:
                    res_lsq[0][i] = abs(res_lsq[0][i])
            res_lsq[0][i]

        for i, k in enumerate(var_est_object.comps['unknown_absorbance']):
            variance_dict[k] = res_lsq[0][i][0]
        variance_dict['device'] = fixed_dev_var
        results.sigma_sq = variance_dict

    return 1


def build_scipy_lsq_arrays(var_est_object):
    """Creates arrays for scipy solvers

    :param VarianceEstimator var_est_object: The variance estimation object

    :return: None

    """
    var_est_object._d_array = np.zeros((len(var_est_object.model.times_spectral), len(var_est_object.model.meas_lambdas)))
    for i, t in enumerate(var_est_object.model.times_spectral):
        for j, l in enumerate(var_est_object.model.meas_lambdas):
            var_est_object._d_array[i, j] = var_est_object.model.D[t, l]

    #if hasattr(var_est_object, '_abs_components'):
    #    n_val = var_est_object._nabs_components
    #else:
    n_val = len(var_est_object.comps['unknown_absorbance'])

    var_est_object._s_array = np.ones(len(var_est_object.model.meas_lambdas) * n_val)
    var_est_object._z_array = np.ones(len(var_est_object.model.times_spectral) * n_val)
    var_est_object._c_array = np.ones(len(var_est_object.model.times_spectral) * n_val)

    return None


def compute_diff_results(results1, results2):
    """Calculate differences between results

    :param results1: The first results object
    :param results2: The second results object

    :return: The ResultsObject containing the differences
    :rtype: ResultsObject

    """
    diff_results = ResultsObject()
    diff_results.Z = results1.Z - results2.Z
    diff_results.S = results1.S - results2.S
    diff_results.C = results1.C - results2.C
    return diff_results


def compute_D_given_SC(var_est_object, results, sigma_d=0):
    """Solves Eq. 23 in Chen et al. 2016

    :param VarianceEstimator var_est_object: The variance estimation object
    :param ResultsObject results: The results object
    :param float sigma_d: Device variance

    :return: None

    """
    d_results = []
    
    for i, t in enumerate(var_est_object.model.times_spectral):
        if t in var_est_object.model.times_spectral:
            for j, l in enumerate(var_est_object.model.meas_lambdas):
                suma = 0.0
                for w, k in enumerate(var_est_object.comps['unknown_absorbance']):
                    Cs = results.C[k][t]  # just the absorbing ones
                    Ss = results.S[k][l]
                    suma += Cs * Ss
                if sigma_d:
                    suma += np.random.normal(0.0, sigma_d)
                d_results.append(suma)

    # if hasattr(var_est_object, '_abs_components'):  # added for removing non_abs ones from first term in obj CS
    #     for i, t in enumerate(var_est_object.model.times_spectral):
    #         if t in var_est_object.model.times_spectral:
    #             for j, l in enumerate(var_est_object.model.meas_lambdas):
    #                 suma = 0.0
    #                 for w, k in enumerate(var_est_object._abs_components):
    #                     Cs = results.C[k][t]  # just the absorbing ones
    #                     Ss = results.S[k][l]
    #                     suma += Cs * Ss
    #                 if sigma_d:
    #                     suma += np.random.normal(0.0, sigma_d)
    #                 d_results.append(suma)

    # else:
    #     for i, t in enumerate(var_est_object.model.times_spectral):
    #         if t in var_est_object.model.times_spectral:
    #             for j, l in enumerate(var_est_object.model.meas_lambdas):
    #                 suma = 0.0
    #                 for w, k in enumerate(var_est_object._mixture_components):
    #                     C = results.C[k][t]
    #                     S = results.S[k][l]
    #                     suma += C * S
    #                 if sigma_d:
    #                     suma += np.random.normal(0.0, sigma_d)
    #                 d_results.append(suma)

    d_array = np.array(d_results).reshape((len(var_est_object.model.times_spectral), len(var_est_object.model.meas_lambdas)))
    results.D = pd.DataFrame(data=d_array,
                             columns=var_est_object.model.meas_lambdas,
                             index=var_est_object.model.times_spectral)

    return None



"""
Reduced Hessian methods renewed.

This module is to hold generic methods for parameter estimator and mee for determining the
reduced hessian. This will eventually replace the reduced hessian module.
"""
# Third party imports
import numpy as np
import pandas as pd
from pathlib import Path
from pyomo.environ import SolverFactory, Suffix
from scipy.sparse import coo_matrix, csr_matrix, csc_matrix, triu

# Kipet library imports
from kipet.general_settings.variable_names import VariableNames
from kipet.model_tools.pyomo_model_tools import convert
    
__var = VariableNames()


def define_free_parameters(models_dict, global_params=None, kind='full'):
    """Identifies the variables in the model used to define the reduced hessian  
       
    :param str kind: full, variable, simple
    
    :return: None
    
    .. todo::
        
        This works with PE too, just supply the r1.p_model
        In MEE, you provide a dict of p_models:
            {k: v.p_model for k, v in self.reaction_models.items()}

    """
    global_params = [] if global_params is None else global_params
    param_names = []
    model_count = False
    
    methods = {'variable': 'getname',
               'simple': 'simple',
               'full': 'to_string'}
    
    if kind not in methods:
        print('Wrong variable naming convention')
    
    def name(method, p, v):
        
        if method in ['to_string', 'getname']:
            return getattr(v, method)()
        else:
            return p
    
    for exp, model in models_dict.items():
        for parameter_kind in __var.optimization_variables:
            if hasattr(model, parameter_kind):
                for p, v in getattr(model, parameter_kind).items():
                    if v.is_fixed() or model_count and v.getname() in global_params:
                        continue
                    param_names.append(name(methods[kind], p, v))
                    
        model_count = True
    
    return param_names


def define_reduce_hess_order(models_dict, component_set, param_names_full):
    """
    This sets up the suffixes of the reduced hessian for use with SIpopt
    
    :param dict models_dict: The dict of parameter estimator models
    :param component_set list: The list of components used
    :param param_names_full: The global parameters as a list (var name)

    :return index_to_variable: Mapping of an index to the model variable
    :rtype: dict

    """
    index_to_variable = dict()
    count_vars = 1
    exp_count = True
    
    for exp, model in models_dict.items():
        if hasattr(model, 'C'):
            var = 'C'
            for index in getattr(model, var).index_set():
                if index[1] in component_set:
                    v = getattr(model, var)[index]
                    index_to_variable[count_vars] = v
                    count_vars += 1
                
    for exp, model in models_dict.items():
        if hasattr(model, 'S') and exp_count:
            var = 'S'
            for index in getattr(model, var).index_set():
                if index[1] in component_set:
                    v = getattr(model, var)[index]
                    index_to_variable[count_vars] = v
                    count_vars += 1
       # exp_count = False
        
    for exp, model in models_dict.items():
        
        for parameter_kind in __var.optimization_variables:
            if hasattr(model, parameter_kind):
                for v in getattr(model, parameter_kind).values():
                    if v.to_string() in param_names_full:
                        index_to_variable[count_vars] = v
                        count_vars += 1

    return index_to_variable


def compute_covariance(models_dict, hessian, free_params, all_variances):
    """This method calculates the covariance matrix for spectral problems
    
    :param dict models_dict: The models in dict form
    :param numpy.ndarray hessian: The reduced hessian
    :param int free_params: The number of free parameters
    :param dict all_variances: The dict of variances (as dict)
    
    :return V_theta: The parameter covariance matrix
    :rtype: scipy.sparse.coo_matrix
    
    """
    H = hessian
    B = make_B_matrix(models_dict, free_params, all_variances)
    Vd = make_Vd_matrix(models_dict, all_variances)
    
    print(f'H: {H.shape = }')
    print(f'B: {B.shape = }')
    print(f'Vd: {Vd.shape = }')
    
    R = B.T @ H.T
    A = Vd @ R
    L = H @ B
    V_theta = 4*(A.T @ L.T).T 
    
    return V_theta
        
    
def make_B_matrix(models_dict, free_params, all_variances):

    """This method generates a B matrix for the covariance calculations in both
    single and multiple experiments.

    :param dict models_dict: The models in dict form
    :param int free_params: The number of free parameters
    
    :return B_matrix: The B matrix
    :rtype: scipy.sparse.coo_matrix
    
    """
    S_part = merge_blocks('S', models_dict, all_variances)
    C_part = merge_blocks('C', models_dict, all_variances)
    
    sy, sx = S_part.shape
    M = np.zeros((C_part.size + S_part.size + free_params, C_part.shape[1]*S_part.shape[1]))
    c_index_start = C_part.size
    
    for i in range(C_part.shape[1]):
    
        M[i*sy:(i + 1)*sy, i*sx:(i + 1)*sx] = S_part
        Ct = C_part[:, i].reshape(-1, 1)
        cy, cx = Ct.shape

        for ci in range(sx):
            M[c_index_start + ci*cy : c_index_start + (ci + 1)*cy, cx*ci + i*sx : cx*ci + i*sx + cx] = Ct
    
    B_matrix = coo_matrix(M)
    
    return B_matrix


def merge_blocks(var, models_dict, all_variances):
    """Merges the defined data types into single blocks
    
    :param str var: The model variable
    
    :return: s_mat: the array of model data
    :rtype: numpy.ndarray
    
    """
    # TODO: Make sure blocks of different sizes (components) can be handled
    s_mat = None
    for e, exp in models_dict.items():
        
        comp_vars = [k for k, v in all_variances[e].items() if k != 'device']    
        
        if s_mat is None:
            s_mat = convert(getattr(exp, var)).loc[:, comp_vars].T.values / all_variances[e]['device']
        else:
            s_mat_append = convert(getattr(exp, var)).loc[:, comp_vars].T.values / all_variances[e]['device']
            s_mat = np.hstack((s_mat, s_mat_append))

    return s_mat


def make_Vd_matrix(models_dict, all_variances):
    """Builds d covariance matrix

    This method is not intended to be used by users directly

    :param dict models_dict: Either a pyomo ConcreteModel or a dict of ReactionModels
    :param dict all_variances: variances

    :return: None

    """
    from kipet.model_tools.pyomo_model_tools import convert
    
    Vd_dict = {}
    M_dict = {}
    total_shape = 0
    n_models = len(models_dict)
    
    for name, model in models_dict.items():

        variances = all_variances[name]
        times = model.allmeas_times.ordered_data()
        waves = model.meas_lambdas.ordered_data()
        n_waves = len(waves)
        n_times = len(times)
        Vd = np.zeros((n_models * n_times * n_waves, n_models * n_times * n_waves))
    
        S = convert(model.S)
        comp_vars = [k for k, v in variances.items() if k != 'device']
        S = S.loc[:, comp_vars]
        
        device_variance = variances['device']
        M = np.array([v for k, v in variances.items() if k != 'device']) * S.values @ S.values.T
        M_diag = np.einsum('ii->i', M)
        M_diag += device_variance
        M_dict[name] = M

        for t in range(n_models*n_times):
            Vd[t*n_waves: (t+1)*n_waves, t*n_waves: (t+1)*n_waves] = M
            
        total_shape += Vd.shape[0]
        Vd_dict[name] = Vd
    
    if n_models > 1:
        
        Vd_combined = np.zeros((total_shape, total_shape))
        start_index = 0
        for model, Vd in Vd_dict.items():        
            Vd_combined[start_index:Vd.shape[0]+start_index, start_index:Vd.shape[1]+start_index] = Vd
            start_index = Vd.shape[0]
        
        return coo_matrix(Vd_combined)
    
    return coo_matrix(Vd)


def index_variable_mapping(model_dict, components, parameter_names, mee_obj=None):
    """This adds the needed suffixes for the reduced hessian to the model object
    used in covariance predictions
    
    :param ConcreteModel model_dict: The model from the parameter estimator or the MEE
    :param list components: The list of absorbing components
    :param list parameter_names: The list of parameter names
    :param mee_obj: An MEE instance, default None
    
    :return: index_to_variables
    :rtype: dict
    
    """
    if mee_obj is not None:
        model_obj = mee_obj
    else:
        model_obj = model_dict
        
    model_obj.red_hessian = Suffix(direction=Suffix.IMPORT_EXPORT)
    index_to_variables = define_reduce_hess_order(model_dict, components, parameter_names)
    
    for k, v in index_to_variables.items():
        model_obj.red_hessian[v] = k
        
    return index_to_variables


def covariance_sipopt(model_obj, solver_factory, components, parameters, mee_obj=None):
    """Generalize the covariance optimization with IPOPT Sens

    :param ConcreteModel model: The Pyomo model used in parameter fitting
    :param SolverFactory optimizer: The SolverFactory currently being used
    :param bool tee: Display option
    :param dict all_sigma_specified: The provided variances
    :param mee_obj: An MEE instance, default None

    :return hessian: The covariance matrix
    :rtype: numpy.ndarray

    """
    from kipet.input_output.read_hessian import split_sipopt_string, read_reduce_hessian
            
    _tmpfile = "ipopt_hess"
    index_to_variable = index_variable_mapping(model_obj, components, parameters, mee_obj)
    
    optimization_model = model_obj if mee_obj is None else mee_obj
    
    solver_results = solver_factory.solve(
        optimization_model,
        tee=False,
        logfile=_tmpfile,
        report_timing=True,
        symbolic_solver_labels=True,
        keepfiles=True)
    
    with open(_tmpfile, 'r') as f:
        output_string = f.read()
    
    ipopt_output, hessian_output = split_sipopt_string(output_string)
    covariance_matrix = read_reduce_hessian(hessian_output, len(index_to_variable))
    covariance_matrix_reduced = covariance_matrix[-len(parameters):, :]
    
    return covariance_matrix, covariance_matrix_reduced


def covariance_k_aug(model_obj, solver_factory, components, parameters, mee_obj=None):
    """Generalize the covariance optimization with IPOPT Sens

    :param ConcreteModel model: The Pyomo model used in parameter fitting
    :param SolverFactory optimizer: The SolverFactory currently being used
    :param bool tee: Display option
    :param dict all_sigma_specified: The provided variances

    :return hessian: The covariance matrix
    :rtype: numpy.ndarray

    """
    _tmpfile = "k_aug_hess"
    optimization_model = model_obj if mee_obj is None else mee_obj
    add_warm_start_suffixes(optimization_model, use_k_aug=True)   
    
    ip = SolverFactory('ipopt')
    solver_results = ip.solve(
        optimization_model,
        tee=False,
        logfile=_tmpfile,
        report_timing=True,
        symbolic_solver_labels=True,
        keepfiles=True
        )

    update_warm_start(optimization_model)
    
    k_aug = SolverFactory('k_aug')
    k_aug.options["print_kkt"] = ""
    k_aug.solve(optimization_model, tee=True)
    stub = ip._problem_files[0][:-3]
    
    var_index_names, con_index_names = var_con_data(stub)
    size = (len(con_index_names), len(var_index_names))
    col_ind, col_ind_param_hr = free_variables(model_obj, components, parameters, var_index_names)
    covariance_matrix = calculate_inverse_hr(size, col_ind)
    covariance_matrix_reduced = covariance_matrix[col_ind_param_hr, :]

    return covariance_matrix, covariance_matrix_reduced


def var_con_data(file_stub):
    """
    This prepares the indecies and sizes for the reduced Hessian strucuturing.
    
    The variables are arranged such that the parameters need to be in the right order
    and entered in last (for k_aug).
    
    The commented code lets you place the variables in any order, as long as they are
    still located at the end (important for B and Vd matrix building.)
    
    """
    var_index = pd.read_csv(Path(file_stub + '.col'), sep=';', header=None)  # dummy sep
    con_index = pd.read_csv(Path(file_stub + '.row'), sep=';', header=None)  # dummy sep
    var_index_names = [var_name for var_name in var_index[0]]
    con_index_names = [con_name for con_name in con_index[0].iloc[:-1]]

    return var_index_names, con_index_names


def free_variables(model, components, parameters, var_index_names): 
    """
    This prepares the indecies and sizes for the reduced Hessian strucuturing.
    
    The variables are arranged such that the parameters need to be in the right order
    and entered in last (for k_aug).
    
    The commented code lets you place the variables in any order, as long as they are
    still located at the end (important for B and Vd matrix building.)
    
    """
    spectral_vars = []
    
    if components is not None:
        if hasattr(model, 'C') and hasattr(model, 'S'):
            for var in ['C', 'S']:
                spectral_vars += [f'{var}[{k[0]},{k[1]}]' for k in getattr(model, var) if k[1] in components]
                
    col_ind = [var_index_names.index(v) for v in spectral_vars + parameters]
    col_ind_P = [var_index_names.index(name) for name in parameters]
    col_ind_param_hr = [col_ind.index(p) for p in col_ind_P]
    
    return col_ind, col_ind_param_hr


def _build_raw_J_and_H(size):
    """Given the size of the variables and constraints, the Hessian and Jacobian
    can be built using the output files from k_aug
    
    :param tuple size: The m (con) and n (var) size of the Jacobian
    
    :return: The Hessian and Jacobian as a tuple of sparse (coo) matrices
    
    """
    m, n = size
    kaug_files = Path('GJH')
    
    hess_file = kaug_files.joinpath('H_print.txt')
    hess = pd.read_csv(hess_file, delim_whitespace=True, header=None, skipinitialspace=True)
    hess.columns = ['irow', 'jcol', 'vals']
    hess.irow -= 1
    hess.jcol -= 1

    jac_file = kaug_files.joinpath('A_print.txt')
    jac = pd.read_csv(jac_file, delim_whitespace=True, header=None, skipinitialspace=True)
    jac.columns = ['irow', 'jcol', 'vals']
    jac.irow -= 1
    jac.jcol -= 1
    
    J = coo_matrix((jac.vals, (jac.jcol, jac.irow)), shape=(m, n))
    Hess_coo = coo_matrix((hess.vals, (hess.irow, hess.jcol)), shape=(n, n))
    H = Hess_coo + triu(Hess_coo, 1).T
    
    return H, J

def _build_reduced_hessian(size, col_ind, con_ind=None, parameter_set=None, delete_fixed_constraints=False):
    """Constructs the reduced Hessian used in various methods.
    
    
    """
    H, J = _build_raw_J_and_H(size)

    if delete_fixed_constraints:
        dummy_constraints = [f'fix_params_to_global[{k}]' for k in parameter_set]
        jac_row_ind = [con_ind.index(d) for d in dummy_constraints]
        # duals_imp = [duals[i] for i in jac_row_ind]

        J_c = delete_from_csr(J.tocsr(), row_indices=jac_row_ind).tocsc()
        row_indexer = SparseRowIndexer(J_c)
        J_f = row_indexer[col_ind]
        J_f = delete_from_csr(J_f.tocsr(), row_indices=jac_row_ind, col_indices=[])
        J_l = delete_from_csr(J_c.tocsr(), col_indices=col_ind)
    
    else:
        J_c = J.tocsc()
        row_indexer = SparseRowIndexer(J_c.T)
        J_f = row_indexer[col_ind].T
        J_l = delete_from_csr(J_c.tocsr(), col_indices=col_ind)
    
    reduced_hessian, Z_mat = _reduced_hessian_matrix(J_f, J_l, H, col_ind)

    return reduced_hessian

def _reduced_hessian_matrix(F, L, H, col_ind):
    """This calculates the reduced hessian by calculating the null-space based
    on the constraints

    :param csr_matrix F: Rows of the Jacobian related to fixed parameters
    :param csr_matrix L: The remainder of the Jacobian without parameters
    :param csr_matrix H: The sparse Hessian
    :param list col_ind: indices of columns with fixed parameters
    
    :return: sparse version of the reduced Hessian
    :rtype: csr_matrix

    """
    from scipy.sparse.linalg import spsolve
    
    n = H.shape[0]
    n_free = n - F.shape[0]
    X = spsolve(L.tocsc(), -F.tocsc())

    col_ind_left = list(set(range(n)).difference(set(col_ind)))
    col_ind_left.sort()

    Z = np.zeros([n, n_free])
    Z[col_ind, :] = np.eye(n_free)

    if isinstance(X, csc_matrix):
        Z[col_ind_left, :] = X.todense()
    else:
        Z[col_ind_left, :] = X.reshape(-1, 1)

    Z_mat = coo_matrix(np.mat(Z)).tocsr()
    Z_mat_T = coo_matrix(np.mat(Z).T).tocsr()
    Hess = H.tocsr()
    reduced_hessian = Z_mat_T * Hess * Z_mat

    return reduced_hessian.todense(), Z_mat


def calculate_inverse_hr(size, col_ind):
    """Calculates the inverse of the reduced Hessian using KKT info (k_aug)
    
    :return H_use: The inverse of the reduced Hessian (parameter rows) 
    :rtype: numpy.ndarray
    
    """
    
    reduced_hessian = _build_reduced_hessian(size, col_ind)
    inv_H_r = np.linalg.inv(reduced_hessian)
    
    return inv_H_r
    
    
def add_warm_start_suffixes(model, use_k_aug=False):
    """Adds suffixed variables to problem

    :param ConcreteModel model: A Pyomo model
    :param bool use_k_aug: Indicates if k_aug solver is being used

    :return: None

    """
    # Ipopt bound multipliers (obtained from solution)
    model.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
    model.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)
    # Ipopt bound multipliers (sent to solver)
    model.ipopt_zL_in = Suffix(direction=Suffix.EXPORT)
    model.ipopt_zU_in = Suffix(direction=Suffix.EXPORT)
    # Obtain dual solutions from first solve and send to warm start
    model.dual = Suffix(direction=Suffix.IMPORT_EXPORT)
    
    if use_k_aug:
        model.dof_v = Suffix(direction=Suffix.EXPORT)
        model.rh_name = Suffix(direction=Suffix.IMPORT)
        
    return None
        
def update_warm_start(model):
    """Updates the suffixed variables for a warmstart

    :param ConcreteModel model: A Pyomo model

    :return: None

    """
    model.ipopt_zL_in.update(model.ipopt_zL_out)
    model.ipopt_zU_in.update(model.ipopt_zU_out)
    
    return None


def calculate_reduced_hessian(model, d=None, optimize=False, parameter_set=[],
                              fix_method=None, rho=10, scaled=True):
    
    if optimize:
        stub = optimize_model(model, 
                              parameter_set=parameter_set, 
                              d=d,
                              fix_method=fix_method,
                              rho=rho,
                              scaled=scaled)
    
    parameter_set_full = define_free_parameters(model, global_params=None, kind='full')
    var_index_names, con_index_names = var_con_data(stub)
    size = (len(con_index_names), len(var_index_names))
    col_ind, col_ind_param_hr = free_variables(model, None, parameter_set_full, var_index_names)
    
    delete_fixed_constraints = fix_method == 'global'
    reduced_hessian = _build_reduced_hessian(size, col_ind, con_index_names, parameter_set, delete_fixed_constraints)
    
    return reduced_hessian

def optimize_model(model, parameter_set=[], d=None, verbose=False, fix_method='fixed',
                    use_bounds=False, variable_name='P', rho=10, scaled=True):
    """Takes the model object and performs the optimization for a reduced Hessian problem.
    
    :param dict d: The current state of the parameters

    :return: None

    """
    from kipet.estimability_tools.parameter_handling import set_scaled_parameter_bounds
    
    if verbose:
        print(f'd: {d}')

    ipopt = SolverFactory('ipopt')
    _tmpfile = 'reduced_hessian'

    if fix_method == 'global':

        if d is not None:
            param_val_dict = {p: d[i] for i, p in enumerate(parameter_set)}
            for k, v in getattr(model, variable_name).items():
                v.set_value(param_val_dict[k])

        add_global_constraints(model, parameter_set, variable_name)

        use_bounds = True
        if use_bounds:
            set_scaled_parameter_bounds(model,
                                        parameter_set=parameter_set,
                                        rho=rho,
                                        scaled=scaled)

    elif fix_method == 'fixed':

        # if hasattr(model, self.global_constraint_name):
        #     self.model_object.del_component(self.global_constraint_name)

        delta = 1e-20
        for k, v in getattr(model, variable_name).items():
            if k in parameter_set:
                ub = v.value
                lb = v.value - delta
                v.setlb(lb)
                v.setub(ub)
                v.unfix()
            else:
                v.fix()
    
    ipopt.solve(model,
                symbolic_solver_labels=True,
                keepfiles=True,
                tee=True,
                logfile=_tmpfile,
                )

    stub = ipopt._problem_files[0][:-3]

    kaug = SolverFactory('k_aug')
    kaug.options["print_kkt"] = ""
    kaug.solve(model, tee=True)

    return stub

def add_global_constraints(model, parameter_set, variable_name):
    """This adds the dummy constraints to the model forcing the local
    parameters to equal the current global parameter values

    :return: None
    
    """
    from pyomo.environ import Constraint, Param, Set
    
    current_set = 'current_p_set'
    global_param_name='d'
    global_constraint_name='fix_params_to_global'
    param_set_name='parameter_names'
    
    if parameter_set is None:
        parameter_set = [p for p in getattr(model, variable_name)]

    # if self.scaled:
    #     global_param_init = {p: 1 for p in parameter_set}
    # else:
    global_param_init = {p: getattr(model, variable_name)[p].value for p in parameter_set}

    for comp in [global_constraint_name, global_param_name, current_set]:
        if hasattr(model, comp):
            model.del_component(comp)

    setattr(model, current_set, Set(initialize=parameter_set))

    setattr(model, global_param_name,
            Param(getattr(model, param_set_name),
                  initialize=global_param_init,
                  mutable=True,
                  ))

    def rule_fix_global_parameters(m, k):
        return getattr(m, variable_name)[k] - getattr(m, global_param_name)[k] == 0

    setattr(model, global_constraint_name,
            Constraint(getattr(model, current_set),
                       rule=rule_fix_global_parameters)
            )

    return None

def delete_from_csr(mat, row_indices=[], col_indices=[]):
    """
    Remove the rows (denoted by ``row_indices``) and columns (denoted by 
    ``col_indices``) from the CSR sparse matrix ``mat``.
    WARNING: Indices of altered axes are reset in the returned matrix

    :param csr_matrix mat: Sparse matrix to delete rows and cols from
    :param list row_indicies: rows to delete
    :param list col_indicies: cols to delete

    :return csr_matrix mat: The sparse matrix with the rows and cols removed

    """
    if not isinstance(mat, csr_matrix):
        raise ValueError("works only for CSR format -- use .tocsr() first")

    rows = []
    cols = []
    if row_indices:
        rows = list(row_indices)
    if col_indices:
        cols = list(col_indices)

    if len(rows) > 0 and len(cols) > 0:
        row_mask = np.ones(mat.shape[0], dtype=bool)
        row_mask[rows] = False
        col_mask = np.ones(mat.shape[1], dtype=bool)
        col_mask[cols] = False
        return mat[row_mask][:, col_mask]
    elif len(rows) > 0:
        mask = np.ones(mat.shape[0], dtype=bool)
        mask[rows] = False
        return mat[mask]
    elif len(cols) > 0:
        mask = np.ones(mat.shape[1], dtype=bool)
        mask[cols] = False
        return mat[:, mask]
    else:
        return mat


class SparseRowIndexer:
    """Class used to splice sparse matrices"""

    def __init__(self, matrix):
        data = []
        indices = []
        indptr = []

        _class = 'csr'
        if isinstance(matrix, csc_matrix):
            _class = 'csc'

        self._class = _class
        # Iterating over the rows this way is significantly more efficient
        # than matrix[row_index,:] and matrix.getrow(row_index)
        for row_start, row_end in zip(matrix.indptr[:-1], matrix.indptr[1:]):
            data.append(matrix.data[row_start:row_end])
            indices.append(matrix.indices[row_start:row_end])
            indptr.append(row_end - row_start)  # nnz of the row

        self.data = np.array(data, dtype=object)
        self.indices = np.array(indices, dtype=object)
        self.indptr = np.array(indptr)
        self.n_columns = matrix.shape[1]

    def __getitem__(self, row_selector):
        data = np.concatenate(self.data[row_selector])
        indices = np.concatenate(self.indices[row_selector])
        indptr = np.append(0, np.cumsum(self.indptr[row_selector]))

        shape = [indptr.shape[0] - 1, self.n_columns]

        if self._class == 'csr':
            return csr_matrix((data, indices, indptr), shape=shape)
        else:
            return csr_matrix((data, indices, indptr), shape=shape).T.tocsc()


"""
The ReactionSet class is a container for ReactionModels when multiple datasets
and model variations are available. This allows the user to perform parameter
fitting with the various models simultaneously.
"""
# Standard library imports
import copy
import inspect
import pathlib
import time

# Third party imports
import pandas as pd

# Kipet library imports
# import kipet.core_methods.data_tools as data_tools
from kipet.estimator_tools.multiple_experiments import MultipleExperimentsEstimator
from kipet.main_modules.reaction_model import ReactionModel
from kipet.general_settings.settings import Settings
from kipet.general_settings.unit_base import UnitBase


class ReactionSet:
    
    """One of two highest level objects in KIPET. This is used to arrange
    multiple models/datasets for simultaneous parameter fitting.
    
    ReactionSet offers many avenues to setting up the problem. 
    
    * You can first generate several separate ReactionModel instances and add
      them to ReactionSet. This can be done using a single module or by importing
      several models from different modules
    
    * You can use the ReactionSet object to create ReactionModel instances and
      then perform parameter fittings with no additional effort
    
    * You can mix both of these methods, if you wish.
    
    Using ReactionSet makes it very simple to generate new ReactionModel
    instances that share many commonalities. This reduces redundant code
    writing if the models are very similar (such as when the only difference 
    is the data and a couple of states).
    
    :var dict reaction_models: The dictionary containing ReactionModel instances
    :var Settings settings: A Settings instance holding all modeling options
    :var dict results: A dictionary of ResultsObjects for each ReactionModel instance
    :var list global_parameters: A list of global parameters
    :var UnitBase ub: UnitBase object used to define the base units of the project
    
    :Methods:
    
    - :func:`add_reaction`
    - :func:`add_reaction_list`
    - :func:`remove_reaction`
    - :func:`new_reaction`
    - :func:`run_opt`
    - :func:`read_data_file`
    - :func:`write_data_file`
    - :func:`add_noise_to_data`
    
    :Properties:
    
    - :func:`all_params`
    
    :Example:
        >>> import kipet
        >>> reaction_lab = kipet.ReactionSet()
    
    """
    def __init__(self):
        """Initialize the ReactionSet instance.
        
        The ReactionSet object does not require any attributes at initialization.
        
        """
        self.reaction_models = {}
        self.settings = Settings()
        self.results = {}
        self.global_parameters = None
        self.ub = UnitBase()
        
        self.file = pathlib.Path(inspect.stack()[1].filename)
        
        t = time.localtime()
        self.timestamp = f'{t.tm_year}-{t.tm_mon:02}-{t.tm_mday:02}-{t.tm_hour:02}-{t.tm_min:02}-{t.tm_sec:02}'

    def __str__(self):
        
        block_str = "ReactionSet\n\n"

        for name, model in self.reaction_models.items():
            block_str += f"{name}\tDatasets: {len(model.datasets)}\n"

        return block_str

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, value):
        return self.reaction_models[value]

    def __len__(self):
        return len(self.reaction_models)

    # def reset_base_units(self):
    #     """Sets the unit base values to the units declared in the settings
        
    #     :return: None
    #     """
    #     self.ub.TIME_BASE = self.settings.units.time
    #     self.ub.VOLUME_BASE = self.settings.units.volume
        
    #     return None


    def add_reaction_list(self, model_list):
        """Method to add a list of ReactionModel instances to the ReactionSet
        
        :parameter list model_list: A list of ReactionModel instances
        
        :return: None
        """
        
        for model in model_list:
            self.add_reaction(model)

        return None


    def remove_reaction(self, model):
        """Remove a model instance from the ReactionSet model list
        
        :parameter str/ReactionModel model: The name or instance of the reaction model to be removed
    
        :return: None
        """
        if isinstance(model, str):
            if model in self.reaction_models:
                self.reaction_models.pop(model)
        elif isinstance(model, ReactionModel):
            self.reaction_models.pop(model.name)
        else:
            print("ReactionSet does not have specified model")
        return None


    def new_reaction(self, name, model=None, ignore=None):

        """Declare new reactions to the ReactionSet using this function

        :parameter str name: The name of the model/experiment used in all references
                         made to it in KIPET, especially in python dicts
        :parameter ReactionModel model: Existing ReactionModel to initialize new ReactionModel with
        :parameter list[str] ignore: This is a list of strings for the various
                attributes in the ReactionModel that should not be copied to the new ReactionModel
                
        - **For example**, if you have a ReactionModel instance (r1) and want to create 
          a new instance (r2) with the same settings except for the data, you could use
          the following::

            r2 = reaction_lab.new_reaction('r2', model=r1, ignore=['datasets'])


        :return: A new ReactionModel instance
            
        """
        ignore = ignore if ignore is not None else []

        if model is None:

            self.reaction_models[name] = ReactionModel(name=name, unit_base=self.ub)
            self.reaction_models[name].file = self.file
            self.reaction_models[name].timestamp = self.timestamp
            
            self.reaction_models[name].settings = self.settings
            
            
            # assign_list = [
            #     "components",
            #     "parameters",
            #     "constants",
            #     "algebraics",
            #     "states",
            #     "ub",
            #     "c",
            # ]

            # for item in assign_list:
            #     if item not in ignore and hasattr(self, item):
            #         setattr(self.reaction_models[name], item, copy.deepcopy(getattr(self, item)))

        else:
            if isinstance(model, ReactionModel):
                
                new_kipet_model = copy.deepcopy(model)
        
                # name = kwargs.get('name', self.name + '_copy')
                # copy_model = kwargs.get('model', True)
                # copy_builder = kwargs.get('builder', True)
                # copy_components = kwargs.get('components', True)   
                # copy_parameters = kwargs.get('parameters', True)
                # copy_datasets = kwargs.get('datasets', True)
                # copy_constants = kwargs.get('constants', True)
                # copy_settings = kwargs.get('settings', True)
                # copy_algebraic_variables = kwargs.get('alg_vars', True)
                # copy_odes = kwargs.get('odes', True)
                # copy_algs = kwargs.get('algs', True)
                
                # Reset the datasets
                
                 #     # Reset the datasets
        
                new_kipet_model.name = name
                
                # if not copy_model:
                #     new_kipet_model.model = None
                
                # if not copy_builder:
                #     new_kipet_model.builder = TemplateBuilder()
                    
                # if not copy_components:
                #     new_kipet_model.components = ComponentBlock()
                
                # if not copy_parameters:
                #     new_kipet_model.parameters = ParameterBlock()
                    
                #if not copy_datasets:
                del new_kipet_model.datasets
                from kipet.model_components.data_component import DataBlock
                new_kipet_model.datasets = DataBlock()
                    
                # if not copy_constants:
                #     new_kipet_model.constants = None
                    
                # if not copy_algebraic_variables:
                #     new_kipet_model.algebraic_variables = []
                    
                # if not copy_settings:
                #     new_kipet_model.settings = Settings()
                    
                # if not copy_odes:
                #     new_kipet_model.odes = None
                    
                # if not copy_algs:
                #     new_kipet_model.algs = None
                
                # list_of_attr_to_delete = ['p_model', 'v_model', 'p_estimator',
                #                           'v_estimator', 'simulator']
                
                # for attr in list_of_attr_to_delete:
                #     if hasattr(new_kipet_model, attr):
                #         setattr(new_kipet_model, attr, None)
                
                # new_kipet_model.results_dict = {}
                
                new_kipet_model.name = name
                self.reaction_models[name] = new_kipet_model
                
            #     kwargs = {}
            #     kwargs["name"] = name
            #     self.reaction_models[name] = ReactionModel(name=name, unit_base=self.ub)
            #     self.reaction_models[name].file = self.file
            #     self.reaction_models[name].timestamp = self.timestamp

            #     assign_list = [
            #         "components",
            #         "parameters",
            #         "constants",
            #         "algebraics",
            #         "states",
            #         "unit_base",
            #         "settings",
            #         "c",
            #         "odes_dict",
            #         "algs_dict",
            #     ]

            #     for item in assign_list:
            #         if item not in ignore and hasattr(model, item):
            #             setattr(self.reaction_models[name], item, copy.deepcopy(getattr(model, item)))

            # else:
            #     raise ValueError("ReactionSet can only add ReactionModel instances.")

        return self.reaction_models[name]
    

    def add_reaction(self, model):
        """Adds a ReactionModel instance to the ReactionSet instance
        
        :parameter ReactionModel model: ReactionModel instance
          
        :return: None
        """
        if isinstance(model, ReactionModel):
            model.unit_base = self.ub
            self.reaction_models[model.name] = model
        else:
            raise ValueError("ReactionSet can only add ReactionModel instances.")

        return None
    
    def run_opt(self, method='mee'):
        """This method will perform parameter fitting for all ReactionModels in
        the ReactionSet models attribute. If more than one ReactionModel instance
        is present, the MultipleExperimentEstiamtor is used to solve for the
        global parameter set.
        
        :parameter str method: Define the method used to solve the multiple
          dataset problem: either **mee** (solves one combined problem simultaneously)
          or **nsd** which solves the problem using a Nested Schur decomposition approach.
        
        .. warning::

            The NSD method is not implemented at this time!
        
        :return: None
        """
        
        if hasattr(self, "mee"):
            del self.mee

        if len(self.reaction_models) > 1:
            if method == "mee":
                self._calculate_parameters()
                self._create_multiple_experiments_estimator()
                from kipet.input_output.kipet_io import Tee

                results_dir = pathlib.Path.cwd().joinpath(self.file.parent, 'results', f'{self.file.stem}-{self.timestamp}')
                if not results_dir.is_dir():
                    results_dir.mkdir(parents=True)

                filename = results_dir.joinpath('log.txt') 

                with Tee(filename): 
                    self._run_full_model()
            # elif method == 'nsd':
            #     self._calculate_parameters()
                # with Tee(f'log-{self.timestamp}.txt'):  
            #         self._mee_nsd(strategy='ipopt')
            else:
                raise ValueError("Not a valid method for optimization")
                
            print('# KIPET procedure for multiple experiments finished\n')

        else:
            reaction_model = self.reaction_models[list(self.reaction_models.keys())[0]]
            results = reaction_model.run_opt()
            self.results[reaction_model.name] = results

        return None

    def _create_multiple_experiments_estimator(self):
        """A quick wrapper for MEE without big changes
        
        """
        
        self.mee = MultipleExperimentsEstimator(self.reaction_models)
        self.mee.confidence_interval = self.settings.parameter_estimator.confidence

        if "spectral" in self.data_types:
            self.settings.general.spectra_problem = True
        else:
            self.settings.general.spectra_problem = False

        self.mee.spectra_problem = self.settings.general.spectra_problem

    def _calculate_parameters(self):
        """Uses the ReactionModel framework to calculate parameters instead of
        repeating this in the MEE

        """
        for name, model in self.reaction_models.items():
            if not model._optimized:
                model.run_opt()
            else:
                print(f"Model {name} has already been optimized")

        return None

    def _run_full_model(self):
        """Runs the MEE after everything has been set up"""
        
        consolidated_settings = self.settings.general
        consolidated_settings.solver_opts = self.settings.solver
        consolidated_settings.covariance = self.settings.parameter_estimator.covariance

        results = self.mee.solve_consolidated_model(
            self.global_parameters, **consolidated_settings
        )

        self.results = results
        for key, results_obj in self.results.items():
            self.reaction_models[key].results = results[key]

        return results

    def _mee_nsd(self, strategy='ipopt'):
        """Performs the NSD on the multiple ReactionModel instances

        :parameter str strategy: Method used to control the outer problem
                ipopt, trust-region, newton-step

        :return results: A dictionary of ResultsObject instances
        :rtype: dict

        """
        kwargs = {'kipet': True,
                  'objective_multiplier': 1
                  }

        if self.global_parameters is not None:
            global_parameters = self.global_parameters
        else:
            global_parameters = self.all_params

        self.nsd = NSD(self.reaction_models,
                        strategy=strategy,
                        global_parameters=global_parameters,
                        kwargs=kwargs)

        #print(self.nsd.d_init)

        results = self.nsd.run_opt()

        self.results = results
        for key, results_obj in self.results.items():
            results_obj.file_dir = self.settings.general.charts_directory

        return results
    

    @property
    def all_params(self):
        """
        Returns the set of all parameters across all ReactionModel instances.
        
        :return: The set of all unique parameters in all models.
        :rtype: set
        """
        
        set_of_all_model_params = set()
        for name, model in self.reaction_models.items():
            set_of_all_model_params = set_of_all_model_params.union(
                model.parameters.names
            )

        return set_of_all_model_params


    @property
    def data_types(self):
        """Returns a set of all data types represented in the contained ReactionModels
        
        :return: A set of data_types as strings
        :rtype: set
        
        """
        data_types = set()
        for name, kipet_model in self.reaction_models.items():
            for dataset_name, dataset in kipet_model.datasets.datasets.items():
                data_types.add(dataset.category)
                
        return data_types
    
    
    @property
    def show_parameters(self):
        """Shows the resulting parameters for each of the ReactionModel instances.
        Naturally, global parameters will have the same values in each model.
         
        :return: DataFrame of the parameters in each reaction
        :rtype: pandas.DataFrame

        """
        df_param = pd.DataFrame(data=None, index=self.all_params, columns=self.reaction_models.keys())
            
        for reaction, model in self.reaction_models.items():
            for param in model.parameters.names:
                df_param.loc[param, reaction] = model.results.P[param]
            
        return df_param
    
    def plot(self, var=None):
        """Plot method for ReactionSet
        
        :param str var: The variable to plot (same as in ReactionModel)
        
        :return: None:
            
        """
        for name, model in self.reaction_models.items():
            
            if var is None:
                model.plot()
            else:
                model.plot(var)
        
    def report(self):
        """Plot method for ReactionSet
        
        :param str var: The variable to plot (same as in ReactionModel)
        
        :return: None:
            
        """
        from kipet.visuals.reports import Report
        for reaction in self.reaction_models.values():
            reaction.plot()
        self.report_object = Report(list(self.reaction_models.values()))
        self.report_object.generate_report()